import asyncio
import json
import logging
import traceback
from typing import Optional

import aiohttp
import aiohttp_cors as aiohttp_cors
import asyncpg.pool
import jwt
import schema
from aiohttp import web
from asyncpg.pool import Pool

from abrechnung.application.accounts import AccountService
from abrechnung.application.groups import GroupService
from abrechnung.application.transactions import TransactionService
from abrechnung.application.users import UserService
from abrechnung.config import Config
from abrechnung.database import db_connect
from abrechnung.http import auth, groups, transactions, websocket, accounts
from abrechnung.http.auth import jwt_middleware, decode_jwt_token
from abrechnung.http.openapi import setup_aiohttp_apispec, validation_middleware
from abrechnung.http.utils import error_middleware, encode_json
from abrechnung.subcommand import SubCommand


class HTTPService(SubCommand):
    """
    sft psql websocket gateway
    """

    def __init__(self, config: Config, **kwargs):
        self.cfg = config

        # map connection_id -> tx queue
        self.tx_queues: dict[int, asyncio.Queue] = dict()

        self.logger = logging.getLogger(__name__)

        # psql notification channel id,
        # we get this when booting from the db.
        self.channel_id: Optional[int] = None
        self.channel_name: Optional[str] = None

    async def run(self):
        """
        run the websocket server
        """

        db_pool = await db_connect(
            username=self.cfg["database"]["user"],
            password=self.cfg["database"]["password"],
            database=self.cfg["database"]["dbname"],
            host=self.cfg["database"]["host"],
        )

        async with db_pool.acquire() as conn:
            # configure automatic decoding of json type postgresql values
            await conn.set_type_codec(
                "json", encoder=json.dumps, decoder=json.loads, schema="pg_catalog"
            )
            await conn.set_type_codec(
                "jsonb", encoder=json.dumps, decoder=json.loads, schema="pg_catalog"
            )

            await self._register_forwarder(conn, forwarder_id=self.cfg["api"]["id"])

            try:
                app = self.create_app(db_pool=db_pool)
                app.router.add_route("GET", "/", self.handle_ws_connection)

                await web._run_app(
                    app, host=self.cfg["api"]["host"], port=self.cfg["api"]["port"]
                )
            finally:
                await self._unregister_forwarder(
                    conn, forwarder_id=self.cfg["api"]["id"]
                )

    async def _register_forwarder(
        self, connection: asyncpg.Connection, forwarder_id: str
    ):
        # register at db
        self.channel_id = await connection.fetchval(
            "select channel_id from forwarder_boot($1)", forwarder_id
        )

        self.logger.info(
            f"Registered forwarder {forwarder_id}: DB gave us channel_id {self.channel_id!r}"
        )
        self.channel_name = f"channel{self.channel_id}"

        # one channel for NOTIFY to this db-client
        await connection.add_listener(self.channel_name, self.on_psql_notification)

        self.logger.info(
            f"Listening on postgresql triggers on channel '{self.channel_name}'"
        )

    async def _unregister_forwarder(
        self, connection: asyncpg.Connection, forwarder_id: str
    ):
        # deregister at db
        await connection.remove_listener(self.channel_name, self.on_psql_notification)

        self.logger.info(f"Unregistered forwarder {forwarder_id}")
        await connection.execute("select * from forwarder_stop($1)", forwarder_id)

    def _create_api_app(
        self, db_pool: Pool, middlewares: Optional[list] = None
    ) -> web.Application:
        api_app = web.Application(middlewares=middlewares)  # type: ignore
        api_app["secret_key"] = self.cfg["api"]["secret_key"]
        api_app["db_pool"] = db_pool
        api_app["config"] = self.cfg

        api_app["user_service"] = UserService(
            db_pool=db_pool,
            config=self.cfg,
        )
        api_app["group_service"] = GroupService(db_pool=db_pool, config=self.cfg)
        api_app["account_service"] = AccountService(db_pool=db_pool, config=self.cfg)
        api_app["transaction_service"] = TransactionService(
            db_pool=db_pool, config=self.cfg
        )

        api_app.add_routes(groups.routes)
        api_app.add_routes(transactions.routes)
        api_app.add_routes(auth.routes)
        api_app.add_routes(accounts.routes)

        api_app.router.add_route("GET", "/ws", self.handle_ws_connection)
        return api_app

    def create_app(
        self,
        db_pool: Pool,
        middlewares: Optional[list] = None,
    ) -> web.Application:
        app = web.Application()
        app["secret_key"] = self.cfg["api"]["secret_key"]
        app["db_pool"] = db_pool
        app["config"] = self.cfg

        if middlewares is None:
            auth_middleware = jwt_middleware(
                secret=self.cfg["api"]["secret_key"],
                whitelist=[
                    "/api/v1/auth/login",
                    "/api/v1/auth/register",
                    "/api/v1/auth/fetch_access_token",
                    "/api/v1/auth/confirm_registration",
                    "/api/v1/auth/confirm_email_change",
                    "/api/v1/auth/recover_password",
                    "/api/v1/auth/confirm_password_recovery",
                    "/api/v1/ws",
                    "/api/v1/docs",
                    "/api/v1/static",
                ],
            )
            middlewares = [auth_middleware]

        middlewares += [error_middleware, validation_middleware]

        api_app = self._create_api_app(db_pool=db_pool, middlewares=middlewares)

        if self.cfg["api"].get("enable_cors", False):
            cors = aiohttp_cors.setup(
                api_app,
                defaults={
                    "*": aiohttp_cors.ResourceOptions(
                        allow_credentials=True,
                        expose_headers="*",
                        allow_headers="*",
                        allow_methods="*",
                    )
                },
            )
            # add all routes to cors exemptions
            for route in list(api_app.router.routes()):
                cors.add(route)

        setup_aiohttp_apispec(
            app=api_app,
            title="Abrechnung OpenAPI Documentation",
            version="v1",
            url="/docs/swagger.json",
        )

        app.add_subapp("/api/v1/", api_app)

        return app

    def on_psql_notification(
        self, connection: asyncpg.Connection, pid: int, channel: str, payload: str
    ):
        """
        this is called by psql to deliver a notify.

        psql sends a json to this notification channel,
        which it knows by the forwarder id assigned to us on boot.

        expected json message from psql:
        {
            -- what connectionids this notification is for
            "connections": [connid, ...],

            -- which event this notification describes
            "event": "...",

            -- event-specific args
            "args": ...
        }

        """
        # del connection, pid  # unused

        self.logger.debug(
            f"Received psql notification on channel {channel} with payload {payload}"
        )

        if channel != self.channel_name:
            raise Exception(
                f"bug: forwarder got a notification "
                f"for channel {channel!r}, "
                f"but registered is {self.channel_name!r}"
            )

        payload_json = json.loads(payload)
        event = payload_json.pop("event")
        connections = payload_json["connections"]
        for connection_id in connections:
            message = {
                "type": "notification",
                "data": {"subscription_type": event, **payload_json["data"]},
            }
            try:
                self.tx_queues[connection_id].put_nowait(message)
            except KeyError:
                pass  # tx queue is no longer available
            except asyncio.QueueFull:
                self.logger.warning(
                    f"[{connection_id}] tx queue full, skipping notification"
                )

    async def handle_ws_connection(self, request):
        """
        how to talk over a websocket connection.
        """
        ws = aiohttp.web.WebSocketResponse()
        await ws.prepare(request)

        async with request.app["db_pool"].acquire() as connection:
            # register the client connection at the db
            connection_id = await connection.fetchval(
                "select connection_id from client_connected($1)", self.channel_id
            )
            self.logger.debug(f"Websocket client connected with id {connection_id}")

        # create the tx queue and task
        tx_queue = asyncio.Queue(maxsize=1000)
        self.tx_queues[connection_id] = tx_queue
        tx_task = asyncio.create_task(self.tx_task(ws, tx_queue))

        try:
            async for msg in ws:
                try:
                    if msg.type == aiohttp.WSMsgType.TEXT:
                        msg_obj = json.loads(msg.data)
                        websocket.CLIENT_SCHEMA.validate(msg_obj)

                        async with request.app["db_pool"].acquire() as connection:
                            response = await self.ws_message(
                                connection, connection_id, msg_obj
                            )
                    else:
                        self.logger.info(
                            f"websocket got unhandled websocket message on connection id {connection_id}: {msg.type}"
                        )
                        continue
                except (
                    asyncpg.DataError,
                    json.JSONDecodeError,
                    schema.SchemaError,
                ) as exc:
                    response = websocket.make_error_msg(
                        code=web.HTTPBadRequest.status_code, msg=str(exc)
                    )
                except Exception as exc:
                    traceback.print_exc()
                    response = websocket.make_error_msg(
                        code=web.HTTPInternalServerError.status_code, msg=str(exc)
                    )
                try:
                    tx_queue.put_nowait(response)
                except asyncio.QueueFull:
                    self.logger.error(
                        f"websocket with id {connection_id} error: tx queue full"
                    )
                    break
        finally:
            async with request.app["db_pool"].acquire() as connection:
                # deregister the client connection
                await connection.execute("call client_disconnected($1)", connection_id)
            # stop the tx task
            del self.tx_queues[connection_id]
            tx_task.cancel()
            await tx_task
            self.logger.debug(f"websocket client with id {connection_id} disconnected")

        return ws

    @staticmethod
    async def tx_task(ws, tx_queue):
        """
        task for sending messages from a queue to a websocket connection
        """
        while True:
            item = await tx_queue.get()
            msg = json.dumps(item, default=encode_json) + "\n"
            await asyncio.shield(ws.send_str(msg))

    async def ws_message(
        self, connection: asyncpg.Connection, connection_id: str, msg: dict
    ):
        """
        the websocket client sent a message. handle it.
        """
        auth_token = msg["token"]
        try:
            jwt_claims = decode_jwt_token(auth_token, self.cfg["api"]["secret_key"])
        except jwt.PyJWTError as exc:
            return websocket.make_error_msg(
                code=web.HTTPUnauthorized.status_code, msg=str(exc)
            )

        msg_type = msg["type"]
        user_id = jwt_claims["user_id"]
        data = msg["data"]
        if msg_type == "subscribe":
            try:
                await connection.execute(
                    "call subscribe($1, $2, $3, $4)",
                    connection_id,
                    user_id,
                    data["subscription_type"],
                    data["element_id"],
                )
                return {"type": "subscribe_success", "data": data}
            except (asyncpg.RaiseError, asyncpg.PostgresError) as exc:
                # a specific error was raised in the db
                return websocket.make_error_msg(
                    code=web.HTTPBadRequest.status_code, msg=str(exc)
                )
        elif msg_type == "unsubscribe":
            try:
                await connection.execute(
                    "call unsubscribe($1, $2, $3, $4)",
                    connection_id,
                    user_id,
                    data["subscription_type"],
                    data["element_id"],
                )
                return {"type": "unsubscribe_success", "data": data}
            except (asyncpg.RaiseError, asyncpg.PostgresError) as exc:
                return websocket.make_error_msg(
                    code=web.HTTPBadRequest.status_code, msg=str(exc)
                )

        else:
            return websocket.make_error_msg(
                code=web.HTTPBadRequest.status_code,
                msg=f"invalid message type {msg_type}",
            )

        raise Exception("reached unreachable place")



from typing import Optional

import asyncpg
from asyncpg.pool import Pool

from abrechnung.config import Config


class NotFoundError(Exception):
    pass


class InvalidCommand(Exception):
    pass


class Application:
    def __init__(self, db_pool: Pool, config: Config):
        self.db_pool = db_pool
        self.cfg = config


async def check_group_permissions(
    conn: asyncpg.Connection,
    group_id: int,
    user_id: int,
    is_owner: bool = False,
    can_write: bool = False,
) -> tuple[bool, bool]:
    membership = await conn.fetchrow(
        "select is_owner, can_write from group_membership where group_id = $1 and user_id = $2",
        group_id,
        user_id,
    )
    if membership is None:
        raise NotFoundError(f"group not found")

    if can_write and not (membership["is_owner"] or membership["can_write"]):
        raise PermissionError(f"write access to group denied")

    if is_owner and not membership["is_owner"]:
        raise PermissionError(f"owner access to group denied")

    return membership["can_write"], membership["is_owner"]


async def create_group_log(
    conn: asyncpg.Connection,
    group_id: int,
    user_id: int,
    type: str,
    message: Optional[str] = None,
    affected_user_id: Optional[int] = None,
):
    await conn.execute(
        "insert into group_log (group_id, user_id, type, message, affected) "
        "values ($1, $2, $3, $4, $5)",
        group_id,
        user_id,
        type,
        "" if message is None else message,
        affected_user_id,
    )



from dataclasses import dataclass
from datetime import datetime


@dataclass
class Session:
    id: int
    name: str
    valid_until: datetime
    last_seen: datetime


@dataclass
class User:
    id: int
    username: str
    email: str
    registered_at: datetime
    deleted: bool
    pending: bool
    sessions: list[Session]



import asyncio
import logging
import os
from unittest import IsolatedAsyncioTestCase as TestCase

import asyncpg
from asyncpg.pool import Pool

from abrechnung.application.users import UserService
from abrechnung.database import revisions

lock = asyncio.Lock()


def get_test_db_config() -> dict:
    return {
        "user": os.environ.get("TEST_DB_USER", "abrechnung-test"),
        "password": os.environ.get("TEST_DB_PASSWORD", "asdf1234"),
        "host": os.environ.get("TEST_DB_HOST", "localhost"),
        "dbname": os.environ.get("TEST_DB_DATABASE", "abrechnung-test"),
        "port": int(os.environ.get("TEST_DB_PORT", 5432)),
    }


async def get_test_db() -> Pool:
    """
    get a connection pool to the test database
    """
    cfg = get_test_db_config()
    pool = await asyncpg.create_pool(
        user=cfg["user"],
        password=cfg["password"],
        database=cfg["dbname"],
        host=cfg["host"],
        port=cfg["port"],
        min_size=5,
        max_size=5,
    )

    await revisions.reset_schema(pool)
    await revisions.apply_revisions(pool)

    return pool


class AsyncTestCase(TestCase):
    def __init__(self, *args, log_level=logging.DEBUG, **kwargs):
        super().__init__(*args, **kwargs)
        logging.basicConfig(level=log_level)

    async def _create_test_user(self, username: str, email: str) -> tuple[int, str]:
        """returns the user id and password"""
        # pylint: disable=protected-access
        async with self.db_pool.acquire() as conn:
            password = "asdf1234"
            hashed_password = UserService._hash_password(password)
            user_id = await conn.fetchval(
                "insert into usr (username, email, hashed_password, pending) values ($1, $2, $3, false) returning id",
                username,
                email,
                hashed_password,
            )

            return user_id, password

    def setUp(self) -> None:
        try:
            self.loop = asyncio.get_running_loop()
        except (AttributeError, RuntimeError):  # AttributeError->py36
            self.loop = asyncio.get_event_loop_policy().get_event_loop()

        self.loop.run_until_complete(self._setup_db())
        self.loop.run_until_complete(self.setUpAsync())

    async def _setup_db(self):
        await lock.acquire()

        self.db_pool = await get_test_db()
        self.db_conn: asyncpg.Connection = await self.db_pool.acquire()

    async def _teardown_db(self):
        await self.db_conn.close()
        await self.db_pool.close()

        lock.release()

    async def setUpAsync(self) -> None:
        pass

    def tearDown(self) -> None:
        self.loop.run_until_complete(self.tearDownAsync())
        self.loop.run_until_complete(self._teardown_db())

    async def tearDownAsync(self) -> None:
        pass



# Configuration file for the Sphinx documentation builder.
#
# This file only contains a selection of the most common options. For a full
# list see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Path setup --------------------------------------------------------------

import datetime

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
#
import json
import sys
from pathlib import Path

HERE = Path(__file__).parent
sys.path[:0] = [str(HERE.parent), str(HERE / "_ext")]
BUILD_DIR = HERE / "_build"

import abrechnung
from abrechnung.config import Config
from abrechnung.http import HTTPService, validation_middleware
from abrechnung.http.openapi import setup_aiohttp_apispec

# -- Project information -----------------------------------------------------

project = "abrechnung"
author = "Michael Loipführer, Jonas Jelten, Michael Enßlin"
copyright = (  # pylint: disable=redefined-builtin
    f"{datetime.datetime.now():%Y}, {author}"
)

version = abrechnung.__version__.replace(".dirty", "")
release = version

# -- General configuration ---------------------------------------------------

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    "sphinx.ext.autodoc",
    "sphinx.ext.intersphinx",
    "sphinx.ext.doctest",
    "sphinx.ext.coverage",
    "sphinx.ext.mathjax",
    "sphinx.ext.napoleon",
    "sphinx.ext.autosummary",
    "sphinx_autodoc_typehints",
    "sphinxcontrib.openapi",
    *[p.stem for p in (HERE / "_ext").glob("*.py")],
]

# Generate the API documentation when building
autosummary_generate = True
autodoc_member_order = "bysource"
napoleon_google_docstring = False
napoleon_numpy_docstring = True
napoleon_include_init_with_doc = False
napoleon_use_rtype = True  # having a separate entry generally helps readability
napoleon_use_param = True
napoleon_custom_sections = [("Params", "Parameters")]
todo_include_todos = False
source_suffix = [".rst", ".md"]

intersphinx_mapping = dict(
    python=("https://docs.python.org/3", None),
)

# Add any paths that contain templates here, relative to this directory.
templates_path = ["_templates"]

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns = ["_build", "Thumbs.db", ".DS_Store"]

# -- Options for HTML output -------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
html_theme = "sphinx_rtd_theme"
html_theme_options = dict(navigation_depth=1, titles_only=True)
github_repo = "abrechnung"
html_context = dict(
    display_github=True,  # Integrate GitHub
    github_user="SFTtech",  # Username
    github_repo="abrechnung",  # Repo name
    github_version="master",  # Version
    conf_py_path="/docs/",  # Path in the checkout to the docs root
)

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
# html_static_path = ["_static"]
html_show_sphinx = False


# generate swagger OpenAPI


def generate_openapi_json():
    config = Config({"api": {"secret_key": "foobar"}})
    service = HTTPService(config)
    app = service._create_api_app(  # pylint: disable=protected-access
        db_pool=None, middlewares=[validation_middleware]
    )  # FIXME: hack to not require db connection

    openapi = setup_aiohttp_apispec(
        app=app,
        title="Abrechnung OpenAPI Documentation",
        version="v1",
        url="/docs/swagger.json",
        in_place=True,
    )
    with open(BUILD_DIR / "openapi.json", "w+", encoding="utf-8") as f:
        json.dump(openapi.swagger_dict(), f)


generate_openapi_json()



from datetime import timezone

from aiohttp import web
from aiohttp.abc import Request
from marshmallow import Schema, fields

from abrechnung.application import NotFoundError
from abrechnung.http.openapi import docs, json_schema
from abrechnung.http.serializers import (
    GroupSchema,
    GroupMemberSchema,
    GroupInviteSchema,
    GroupPreviewSchema,
    GroupLogSchema,
)
from abrechnung.http.utils import json_response

routes = web.RouteTableDef()


@routes.get("/groups")
@docs(tags=["groups"], summary="list the current users groups", description="")
async def list_groups(request):
    try:
        groups = await request.app["group_service"].list_groups(
            user_id=request["user"]["user_id"]
        )
    except NotFoundError:
        raise web.HTTPForbidden(reason="permission denied")

    serializer = GroupSchema()

    return json_response(data=serializer.dump(groups, many=True))


@routes.post("/groups")
@docs(tags=["groups"], summary="create a group", description="")
@json_schema(
    Schema.from_dict(
        {
            "name": fields.Str(),
            "description": fields.Str(),
            "currency_symbol": fields.Str(),
            "terms": fields.Str(),
        },
        name="CreateGroupSchema",
    )
)
async def create_group(request: Request):
    data = request["json"]
    group_id = await request.app["group_service"].create_group(
        user_id=request["user"]["user_id"],
        name=data["name"],
        description=data["description"],
        currency_symbol=data["currency_symbol"],
        terms=data["terms"],
    )

    return json_response(data={"group_id": group_id})


@routes.get(r"/groups/{group_id:\d+}")
@docs(tags=["groups"], summary="fetch group details", description="")
async def get_group(request: Request):
    group = await request.app["group_service"].get_group(
        user_id=request["user"]["user_id"],
        group_id=int(request.match_info["group_id"]),
    )

    serializer = GroupSchema()

    return json_response(data=serializer.dump(group))


@routes.post(r"/groups/{group_id:\d+}")
@docs(tags=["groups"], summary="update group details", description="")
@json_schema(
    Schema.from_dict(
        {
            "name": fields.Str(),
            "description": fields.Str(),
            "currency_symbol": fields.Str(),
            "terms": fields.Str(),
        },
        name="UpdateGroupSchema",
    )
)
async def update_group(request: Request):
    data = request["json"]
    await request.app["group_service"].update_group(
        user_id=request["user"]["user_id"],
        group_id=int(request.match_info["group_id"]),
        name=data["name"],
        description=data["description"],
        currency_symbol=data["currency_symbol"],
        terms=data["terms"],
    )

    return web.Response(status=web.HTTPNoContent.status_code)


@routes.delete(r"/groups/{group_id:\d+}")
@docs(tags=["groups"], summary="delete a group", description="")
async def delete_group(request: Request):
    await request.app["group_service"].delete_group(
        user_id=request["user"]["user_id"],
        group_id=int(request.match_info["group_id"]),
    )

    return web.Response(status=web.HTTPNoContent.status_code)


@routes.post(r"/groups/{group_id:\d+}/leave")
@docs(tags=["groups"], summary="leave a group", description="")
async def leave_group(request: Request):
    await request.app["group_service"].leave_group(
        user_id=request["user"]["user_id"],
        group_id=int(request.match_info["group_id"]),
    )

    return web.Response(status=web.HTTPNoContent.status_code)


@routes.get(r"/groups/{group_id:\d+}/members")
@docs(tags=["groups"], summary="list all members of a group", description="")
async def list_members(request: web.Request):
    members = await request.app["group_service"].list_members(
        user_id=request["user"]["user_id"],
        group_id=int(request.match_info["group_id"]),
    )

    serializer = GroupMemberSchema()

    return json_response(data=serializer.dump(members, many=True))


@routes.get(r"/groups/{group_id:\d+}/logs")
@docs(tags=["groups"], summary="fetch the group log", description="")
async def list_log(request: web.Request):
    logs = await request.app["group_service"].list_log(
        user_id=request["user"]["user_id"],
        group_id=int(request.match_info["group_id"]),
    )

    serializer = GroupLogSchema()

    return json_response(data=serializer.dump(logs, many=True))


@routes.post(r"/groups/{group_id:\d+}/send_message")
@docs(tags=["groups"], summary="post a message to the group log", description="")
@json_schema(Schema.from_dict({"message": fields.Str()}, name="SendGroupMessageSchema"))
async def send_group_message(request: web.Request):
    data = request["json"]
    await request.app["group_service"].send_group_message(
        user_id=request["user"]["user_id"],
        group_id=int(request.match_info["group_id"]),
        message=data["message"],
    )

    return web.Response(status=web.HTTPNoContent.status_code)


@routes.post(r"/groups/{group_id:\d+}/members")
@docs(
    tags=["groups"], summary="update the permissions of a group member", description=""
)
@json_schema(
    Schema.from_dict(
        {"user_id": fields.Int(), "can_write": fields.Bool(), "is_owner": fields.Bool()}
    ),
    name="UpdateMemberPermissionsSchema",
)
async def update_member_permissions(request: web.Request):
    data = request["json"]
    await request.app["group_service"].update_member_permissions(
        user_id=request["user"]["user_id"],
        group_id=int(request.match_info["group_id"]),
        member_id=data["user_id"],
        can_write=data["can_write"],
        is_owner=data["is_owner"],
    )

    return web.Response(status=web.HTTPNoContent.status_code)


@routes.get(r"/groups/{group_id:\d+}/invites")
@docs(tags=["groups"], summary="list all invite links of a group", description="")
async def list_invites(request):
    invites = await request.app["group_service"].list_invites(
        user_id=request["user"]["user_id"],
        group_id=int(request.match_info["group_id"]),
    )

    serializer = GroupInviteSchema()

    return json_response(data=serializer.dump(invites, many=True))


@routes.post(r"/groups/{group_id:\d+}/invites")
@docs(tags=["groups"], summary="create a new group invite link", description="")
@json_schema(
    Schema.from_dict(
        {
            "description": fields.Str(),
            "single_use": fields.Bool(),
            "join_as_editor": fields.Bool(),
            "valid_until": fields.DateTime(),
        },
        name="CreateInviteSchema",
    )
)
async def create_invite(request: Request):
    data = request["json"]
    valid_until = data["valid_until"]
    if valid_until.tzinfo is None:
        valid_until = valid_until.replace(tzinfo=timezone.utc)

    await request.app["group_service"].create_invite(
        user_id=request["user"]["user_id"],
        group_id=int(request.match_info["group_id"]),
        description=data["description"],
        single_use=data["single_use"],
        valid_until=valid_until,
        join_as_editor=data["join_as_editor"],
    )

    return json_response(status=web.HTTPNoContent.status_code)


@routes.delete(r"/groups/{group_id:\d+}/invites/{invite_id:\d+}")
@docs(tags=["groups"], summary="delete a group invite link", description="")
async def delete_invite(request: Request):
    await request.app["group_service"].delete_invite(
        user_id=request["user"]["user_id"],
        group_id=int(request.match_info["group_id"]),
        invite_id=int(request.match_info["invite_id"]),
    )
    return json_response(status=web.HTTPNoContent.status_code)


@routes.post(r"/groups/preview")
@docs(
    tags=["groups"],
    summary="preview a group before joining using an invite token",
    description="",
)
@json_schema(
    Schema.from_dict({"invite_token": fields.Str()}, name="PreviewGroupSchema")
)
async def preview_group(request: Request):
    data = request["json"]
    group_preview = await request.app["group_service"].preview_group(
        invite_token=data["invite_token"],
    )

    serializer = GroupPreviewSchema()

    return json_response(data=serializer.dump(group_preview))


@routes.post(r"/groups/join")
@docs(tags=["groups"], summary="join a group using an invite token", description="")
@json_schema(Schema.from_dict({"invite_token": fields.Str()}, name="JoinGroupSchema"))
async def join_group(request: Request):
    data = request["json"]
    await request.app["group_service"].join_group(
        user_id=request["user"]["user_id"],
        invite_token=data["invite_token"],
    )

    return json_response(status=web.HTTPNoContent.status_code)



"""
Loosely based on gist.github.com/MantasVaitkunas/7c16de233812adcb7028
This version is taken from https://github.com/theislab/scCODA/blob/master/docs/source/_ext/edit_on_github.py
"""

import warnings

__licence__ = "BSD (3 clause)"


def get_github_repo(app):
    return app.config.github_repo, "/docs/"


def html_page_context(app, pagename, templatename, context, doctree):
    if templatename != "page.html":
        return

    if not app.config.github_repo:
        warnings.warn("`github_repo `not specified")
        return

    repo, conf_py_path = get_github_repo(app)

    # For sphinx_rtd_theme.
    context["display_github"] = True
    context["github_user"] = "SFTtech"
    context["github_version"] = "master"
    context["github_repo"] = repo
    context["conf_py_path"] = conf_py_path


def setup(app):
    app.add_config_value("github_repo", "", True)
    app.connect("html-page-context", html_page_context)



from marshmallow import Schema, fields
from aiohttp import web

from abrechnung.http.openapi import docs, json_schema
from abrechnung.http.serializers import AccountSchema
from abrechnung.http.utils import json_response

routes = web.RouteTableDef()


async def _account_response(request, account_id: int) -> web.Response:
    account = await request.app["account_service"].get_account(
        user_id=request["user"]["user_id"], account_id=account_id
    )

    serializer = AccountSchema()
    return json_response(data=serializer.dump(account))


@routes.get(r"/groups/{group_id:\d+}/accounts")
@docs(
    tags=["accounts"],
    summary="list all accounts in a group",
    description="",
    responses={200: {"schema": AccountSchema(many=True), "description": ""}},
)
async def list_accounts(request):
    try:
        accounts = await request.app["account_service"].list_accounts(
            user_id=request["user"]["user_id"],
            group_id=int(request.match_info["group_id"]),
        )
    except PermissionError:
        raise web.HTTPForbidden(reason="permission denied")

    serializer = AccountSchema()
    return json_response(data=serializer.dump(accounts, many=True))


@routes.post(r"/groups/{group_id:\d+}/accounts")
@docs(
    tags=["accounts"],
    summary="create a new group account",
    description="",
    responses={200: {"schema": AccountSchema(), "description": ""}},
)
@json_schema(
    Schema.from_dict(
        {"name": fields.Str(), "description": fields.Str(), "type": fields.Str()},
        name="CreateAccountSchema",
    )
)
async def create_account(request: web.Request):
    data = request["json"]
    account_id = await request.app["account_service"].create_account(
        user_id=request["user"]["user_id"],
        group_id=int(request.match_info["group_id"]),
        name=data["name"],
        description=data["description"],
        type=data["type"],
    )

    return await _account_response(request=request, account_id=account_id)


@routes.get(r"/accounts/{account_id:\d+}")
@docs(
    tags=["accounts"],
    summary="fetch a group account",
    description="",
    responses={200: {"schema": AccountSchema(), "description": ""}},
)
async def get_account(request: web.Request):
    account_id = int(request.match_info["account_id"])
    return await _account_response(request=request, account_id=account_id)


@routes.post(r"/accounts/{account_id:\d+}")
@docs(
    tags=["accounts"],
    summary="update an account",
    description="",
    responses={200: {"schema": AccountSchema(), "description": ""}},
)
@json_schema(
    Schema.from_dict(
        {"name": fields.Str(), "description": fields.Str()}, name="UpdateAccountSchema"
    )
)
async def update_account(request: web.Request):
    account_id = int(request.match_info["account_id"])
    data = request["json"]
    await request.app["account_service"].update_account(
        user_id=request["user"]["user_id"],
        account_id=account_id,
        name=data["name"],
        description=data["description"],
    )

    return await _account_response(request=request, account_id=account_id)


@routes.delete(r"/accounts/{account_id:\d+}")
@docs(
    tags=["accounts"],
    summary="delete an account",
    description="",
    responses={200: {"schema": AccountSchema(), "description": ""}},
)
async def delete_account(request: web.Request):
    account_id = int(request.match_info["account_id"])
    await request.app["account_service"].delete_account(
        user_id=request["user"]["user_id"],
        account_id=account_id,
    )

    return await _account_response(request=request, account_id=account_id)



from datetime import datetime, timezone
from typing import Optional, Union

import asyncpg

from abrechnung.domain.accounts import Account
from . import (
    Application,
    NotFoundError,
    check_group_permissions,
    InvalidCommand,
    create_group_log,
)


class AccountService(Application):
    @staticmethod
    async def _increment_account_version(conn: asyncpg.Connection, revision_id: int):
        await conn.execute(
            "update account_revision set version = version + 1 where id = $1",
            revision_id,
        )

    @staticmethod
    async def _check_account_permissions(
        conn: asyncpg.Connection,
        user_id: int,
        account_id: int,
        can_write: bool = False,
        account_type: Optional[Union[str, list[str]]] = None,
    ) -> int:
        """returns group id of the transaction"""
        result = await conn.fetchrow(
            "select a.type, a.group_id, can_write, is_owner "
            "from group_membership gm join account a on gm.group_id = a.group_id and gm.user_id = $1 "
            "where a.id = $2",
            user_id,
            account_id,
        )
        if not result:
            raise NotFoundError(f"account not found")

        if can_write and not (result["can_write"] or result["is_owner"]):
            raise PermissionError(f"user does not have write permissions")

        if account_type:
            type_check = (
                [account_type] if isinstance(account_type, str) else account_type
            )
            if result["type"] not in type_check:
                raise InvalidCommand(
                    f"Transaction type {result['type']} does not match the expected type {type_check}"
                )

        return result["group_id"]

    async def list_accounts(self, *, user_id: int, group_id: int) -> list[Account]:
        async with self.db_pool.acquire() as conn:
            async with conn.transaction():
                await check_group_permissions(
                    conn=conn, group_id=group_id, user_id=user_id
                )
                cur = conn.cursor(
                    "select account_id, group_id, type, revision_id, name, description, priority, deleted "
                    "from committed_account_state_valid_at() "
                    "where group_id = $1",
                    group_id,
                )
                result = []
                async for account in cur:
                    result.append(
                        Account(
                            id=account["account_id"],
                            group_id=account["group_id"],
                            type=account["type"],
                            name=account["name"],
                            description=account["description"],
                            priority=account["priority"],
                            deleted=account["deleted"],
                        )
                    )

                return result

    async def get_account(self, *, user_id: int, account_id: int) -> Account:
        async with self.db_pool.acquire() as conn:
            await self._check_account_permissions(
                conn=conn, user_id=user_id, account_id=account_id
            )
            account = await conn.fetchrow(
                "select account_id, group_id, type, revision_id, name, description, priority, deleted "
                "from committed_account_state_valid_at() "
                "where account_id = $1",
                account_id,
            )
            if account is None:
                raise NotFoundError(f"No account with id {account_id} exists")

            return Account(
                id=account["account_id"],
                group_id=account["group_id"],
                type=account["type"],
                name=account["name"],
                description=account["description"],
                priority=account["priority"],
                deleted=account["deleted"],
            )

    async def create_account(
        self,
        *,
        user_id: int,
        group_id: int,
        type: str,
        name: str,
        description: str,
        priority: int = 0,
    ) -> int:
        async with self.db_pool.acquire() as conn:
            async with conn.transaction():
                await check_group_permissions(
                    conn=conn, group_id=group_id, user_id=user_id, can_write=True
                )
                account_id = await conn.fetchval(
                    "insert into account (group_id, type) values ($1, $2) returning id",
                    group_id,
                    type,
                )
                now = datetime.now(tz=timezone.utc)
                revision_id = await conn.fetchval(
                    "insert into account_revision (user_id, account_id, started, committed) "
                    "values ($1, $2, $3, $4) returning id",
                    user_id,
                    account_id,
                    now,
                    now,
                )
                await conn.execute(
                    "insert into account_history (id, revision_id, name, description, priority) "
                    "values ($1, $2, $3, $4, $5)",
                    account_id,
                    revision_id,
                    name,
                    description,
                    priority,
                )
                await create_group_log(
                    conn=conn,
                    group_id=group_id,
                    user_id=user_id,
                    type="account-committed",
                    message=f"created account {name}",
                )
                return account_id

    async def update_account(
        self,
        user_id: int,
        account_id: int,
        name: str,
        description: str,
        priority: int = 0,
    ):
        # TODO: figure out the more complex logic once we have accounts stuck in uncommitted states
        async with self.db_pool.acquire() as conn:
            async with conn.transaction():
                group_id = await self._check_account_permissions(
                    conn=conn, user_id=user_id, account_id=account_id, can_write=True
                )
                account = await conn.fetchrow(
                    "select account_id, type, revision_id, name, description, priority "
                    "from committed_account_state_valid_at() "
                    "where account_id = $1 and deleted = false",
                    account_id,
                )
                if account is None:
                    raise NotFoundError(f"No account with id {account_id} exists")

                if (
                    name != account["name"]
                    or description != account["description"]
                    or priority != account["priority"]
                ):
                    """if there is something to change initialize a new revision and a new entry in the history table"""
                    now = datetime.now(tz=timezone.utc)
                    revision_id = await conn.fetchval(
                        "insert into account_revision (user_id, account_id, started, committed) "
                        "values ($1, $2, $3, $4) returning id",
                        user_id,
                        account_id,
                        now,
                        now,
                    )
                    await conn.execute(
                        "insert into account_history (id, revision_id, name, description, priority) "
                        "values ($1, $2, $3, $4, $5)",
                        account_id,
                        revision_id,
                        name,
                        description,
                        priority,
                    )
                    await create_group_log(
                        conn=conn,
                        group_id=group_id,
                        user_id=user_id,
                        type="account-committed",
                        message=f"updated account {name}",
                    )

    async def delete_account(
        self,
        user_id: int,
        account_id: int,
    ):
        async with self.db_pool.acquire() as conn:
            async with conn.transaction():
                group_id = await self._check_account_permissions(
                    conn=conn, user_id=user_id, account_id=account_id, can_write=True
                )
                row = await conn.fetchrow(
                    "select id from account where id = $1",
                    account_id,
                )
                if row is None:
                    raise InvalidCommand(f"Account does not exist")

                has_committed_shares = await conn.fetchval(
                    "select 1 "
                    "from committed_transaction_state_valid_at() t "
                    "where not deleted and $1 = any(involved_accounts)",
                    account_id,
                )
                has_pending_shares = await conn.fetchval(
                    "select 1 "
                    "from aggregated_pending_transaction_history t "
                    "where $1 = any(t.involved_accounts)",
                    account_id,
                )

                has_committed_usages = await conn.fetchval(
                    "select 1 "
                    "from committed_transaction_position_state_valid_at() p "
                    "join transaction t on t.id = p.transaction_id "
                    "where not p.deleted and $1 = any(p.involved_accounts)",
                    account_id,
                )

                has_pending_usages = await conn.fetchval(
                    "select 1 "
                    "from aggregated_pending_transaction_position_history p "
                    "join transaction t on t.id = p.transaction_id "
                    "where $1 = any(p.involved_accounts)",
                    account_id,
                )

                if (
                    has_committed_shares
                    or has_pending_shares
                    or has_committed_usages
                    or has_pending_usages
                ):
                    raise InvalidCommand(
                        f"Cannot delete an account that is references by a transaction"
                    )

                row = await conn.fetchrow(
                    "select name, revision_id, deleted "
                    "from committed_account_state_valid_at() "
                    "where account_id = $1",
                    account_id,
                )
                if row is None:
                    raise InvalidCommand(
                        f"Cannot delete an account without any committed changes"
                    )

                if row["deleted"]:
                    raise InvalidCommand(f"Cannot delete an already deleted account")

                now = datetime.now(tz=timezone.utc)
                revision_id = await conn.fetchval(
                    "insert into account_revision (user_id, account_id, started, committed) "
                    "values ($1, $2, $3, $4) returning id",
                    user_id,
                    account_id,
                    now,
                    now,
                )
                await conn.execute(
                    "insert into account_history (id, revision_id, name, description, priority, deleted) "
                    "select $1, $2, name, description, priority, true "
                    "from account_history ah where ah.id = $1 and ah.revision_id = $3 ",
                    account_id,
                    revision_id,
                    row["revision_id"],
                )

                await create_group_log(
                    conn=conn,
                    group_id=group_id,
                    user_id=user_id,
                    type="account-deleted",
                    message=f"deleted account account {row['name']}",
                )



import asyncio
import json
from datetime import datetime, timedelta, date, timezone

import aiohttp
from aiohttp import web

from abrechnung.http.auth import token_for_user
from tests.http_tests import BaseHTTPAPITest


class WebsocketAPITest(BaseHTTPAPITest):
    async def expect_ws_message(self, ws: web.WebSocketResponse, expected_msg: dict):
        resp = await ws.receive(timeout=1)
        self.assertEqual(resp.type, aiohttp.WSMsgType.TEXT)
        resp_json = json.loads(resp.data)
        self.assertEqual(expected_msg, resp_json | expected_msg)

    async def test_websocket_notifications(self):
        user_id, password = await self._create_test_user(
            username="user", email="email@email.com"
        )
        _, session_id, _ = await self.user_service.login_user(
            "user", password=password, session_name="session1"
        )
        token = token_for_user(
            user_id, session_id=session_id, secret_key=self.secret_key
        )

        ws = await self.client.ws_connect("/api/v1/ws")
        self.assertIsNotNone(ws)

        await ws.send_json(
            {
                "type": "subscribe",
                "token": token,
                "data": {"subscription_type": "group", "element_id": user_id},
            }
        )

        await self.expect_ws_message(
            ws,
            {
                "type": "subscribe_success",
                "data": {
                    "element_id": user_id,
                    "subscription_type": "group",
                },
            },
        )

        group_id = await self.group_service.create_group(
            user_id=user_id,
            name="group1",
            description="asdf",
            currency_symbol="€",
            terms="terms...",
        )

        # make sure we receive a notification on the group scope
        await self.expect_ws_message(
            ws,
            {
                "type": "notification",
                "data": {
                    "element_id": user_id,
                    "group_id": group_id,
                    "subscription_type": "group",
                },
            },
        )

        await ws.send_json(
            {
                "type": "subscribe",
                "token": token,
                "data": {"subscription_type": "account", "element_id": group_id},
            }
        )

        await self.expect_ws_message(
            ws,
            {
                "type": "subscribe_success",
                "data": {
                    "element_id": group_id,
                    "subscription_type": "account",
                },
            },
        )

        account_id = await self.account_service.create_account(
            user_id=user_id,
            group_id=group_id,
            type="personal",
            name="group1",
            description="asdf",
        )
        await self.expect_ws_message(
            ws,
            {
                "type": "notification",
                "data": {
                    "element_id": group_id,
                    "account_id": account_id,
                    "subscription_type": "account",
                },
            },
        )

        # subscribe to transaction changes
        await ws.send_json(
            {
                "type": "subscribe",
                "token": token,
                "data": {"subscription_type": "transaction", "element_id": group_id},
            }
        )

        await self.expect_ws_message(
            ws,
            {
                "type": "subscribe_success",
                "data": {
                    "element_id": group_id,
                    "subscription_type": "transaction",
                },
            },
        )

        transaction_id = await self.transaction_service.create_transaction(
            user_id=user_id,
            group_id=group_id,
            type="transfer",
            value=1.2,
            currency_symbol="€",
            currency_conversion_rate=1.0,
            billed_at=date.today(),
            description="asdf",
        )

        await self.expect_ws_message(
            ws,
            {
                "type": "notification",
                "data": {
                    "element_id": group_id,
                    "transaction_id": transaction_id,
                    "subscription_type": "transaction",
                    "deleted": False,
                    "revision_committed": None,
                    "revision_version": 0,
                },
            },
        )

        # subscribe to group invite changes
        await ws.send_json(
            {
                "type": "subscribe",
                "token": token,
                "data": {"subscription_type": "group_invite", "element_id": group_id},
            }
        )

        await self.expect_ws_message(
            ws,
            {
                "type": "subscribe_success",
                "data": {
                    "element_id": group_id,
                    "subscription_type": "group_invite",
                },
            },
        )

        invite_id = await self.group_service.create_invite(
            user_id=user_id,
            group_id=group_id,
            description="foobar invite",
            single_use=False,
            valid_until=datetime.now() + timedelta(hours=4),
            join_as_editor=True,
        )

        await self.expect_ws_message(
            ws,
            {
                "type": "notification",
                "data": {
                    "element_id": group_id,
                    "invite_id": invite_id,
                    "subscription_type": "group_invite",
                },
            },
        )

        await ws.send_json(
            {
                "type": "unsubscribe",
                "token": token,
                "data": {"subscription_type": "transaction", "element_id": group_id},
            }
        )

        await self.expect_ws_message(
            ws,
            {
                "type": "unsubscribe_success",
                "data": {
                    "element_id": group_id,
                    "subscription_type": "transaction",
                },
            },
        )

        await ws.send_json(
            {
                "type": "unsubscribe",
                "token": token,
                "data": {"subscription_type": "account", "element_id": group_id},
            }
        )

        await self.expect_ws_message(
            ws,
            {
                "type": "unsubscribe_success",
                "data": {
                    "element_id": group_id,
                    "subscription_type": "account",
                },
            },
        )

        await ws.send_json(
            {
                "type": "unsubscribe",
                "token": token,
                "data": {"subscription_type": "group", "element_id": user_id},
            }
        )

        await self.expect_ws_message(
            ws,
            {
                "type": "unsubscribe_success",
                "data": {
                    "element_id": user_id,
                    "subscription_type": "group",
                },
            },
        )



import logging
import re
from datetime import timedelta, datetime, timezone

import jwt
from aiohttp import web, hdrs
from marshmallow import Schema, fields

from abrechnung.application import InvalidCommand
from abrechnung.application.users import (
    InvalidPassword,
)
from abrechnung.http.serializers import UserSchema
from abrechnung.http.utils import json_response
from abrechnung.http.openapi import docs, json_schema

logger = logging.getLogger(__name__)

routes = web.RouteTableDef()

REQUEST_AUTH_KEY = "user"
ACCESS_TOKEN_VALIDITY = timedelta(hours=1)


def check_request(request, entries):
    for pattern in entries:
        if re.match(pattern, request.path):
            return True

    return False


def access_token_expiry() -> datetime:
    return datetime.now(tz=timezone.utc) + ACCESS_TOKEN_VALIDITY


def token_for_user(user_id: int, session_id: int, secret_key: str) -> str:
    return jwt.encode(
        {"exp": access_token_expiry(), "user_id": user_id, "session_id": session_id},
        secret_key,
        algorithm="HS256",
    )


def decode_jwt_token(token: str, secret: str) -> dict:
    return jwt.decode(token, secret, algorithms=["HS256"])


def jwt_middleware(
    secret,
    whitelist=tuple(),
    auth_scheme="Bearer",
):
    """Mostly taken from https://github.com/hzlmn/aiohttp-jwt"""
    if not (secret and isinstance(secret, str)):
        raise RuntimeError(
            "secret or public key should be provided for correct work",
        )

    @web.middleware
    async def _jwt_middleware(request, handler):
        if request.method == hdrs.METH_OPTIONS:
            return await handler(request)

        if check_request(request, whitelist):
            return await handler(request)

        token = None

        if "Authorization" in request.headers:
            try:
                scheme, token = request.headers.get("Authorization").strip().split(" ")
            except ValueError:
                raise web.HTTPForbidden(reason="Invalid authorization header")

            if not re.match(auth_scheme, scheme):
                raise web.HTTPForbidden(reason="Invalid token scheme")

        if token is None:
            raise web.HTTPUnauthorized(
                reason="Missing authorization token",
            )

        if not isinstance(token, bytes):
            token = token.encode()

        try:
            decoded = decode_jwt_token(token, secret)
        except jwt.PyJWTError as exc:
            logger.info(f"Received invalid authorization token: {exc}")
            msg = "Invalid authorization token, " + str(exc)
            raise web.HTTPUnauthorized(reason=msg)

        if "user_id" not in decoded or "session_id" not in decoded:
            raise web.HTTPUnauthorized(reason="Invalid token claims")

        async with request.app["db_pool"].acquire() as conn:
            session_id = await conn.fetchval(
                "select id from session where id = $1 and user_id = $2 and valid_until is null or valid_until > now()",
                decoded["session_id"],
                decoded["user_id"],
            )
            if not session_id:
                raise web.HTTPUnauthorized(
                    reason="provided access token for expired or logged out session"
                )

        request[REQUEST_AUTH_KEY] = {
            "user_id": decoded["user_id"],
            "session_id": decoded["session_id"],
        }

        return await handler(request)

    return _jwt_middleware


@routes.post("/auth/login")
@docs(tags=["auth"], summary="login with username and password", description="")
@json_schema(
    Schema.from_dict(
        {
            "username": fields.Str(),
            "password": fields.Str(),
            "session_name": fields.Str(),
        },
        name="LoginSchema",
    )
)
async def login(request):
    data = request["json"]
    user_id, session_id, session_token = await request.app["user_service"].login_user(
        username=data["username"],
        password=data["password"],
        session_name=data["session_name"],
    )

    token = token_for_user(
        user_id=user_id, session_id=session_id, secret_key=request.app["secret_key"]
    )

    return json_response(
        data={
            "user_id": user_id,
            "access_token": token,
            "session_token": session_token,
        }
    )


@routes.post("/auth/logout")
@docs(tags=["auth"], summary="sign out of the current session", description="")
async def logout(request):
    await request.app["user_service"].logout_user(
        session_id=request["user"]["session_id"], user_id=request["user"]["user_id"]
    )
    return web.Response(status=web.HTTPNoContent.status_code)


@routes.post("/auth/fetch_access_token")
@docs(
    tags=["auth"],
    summary="get a short lived access token ussing a session token",
    description="",
)
@json_schema(Schema.from_dict({"token": fields.Str()}, name="FetchAccessTokenSchema"))
async def fetch_access_token(request):
    data = request["json"]
    row = await request.app["user_service"].is_session_token_valid(token=data["token"])
    if row is None:
        raise web.HTTPBadRequest(reason="invalid session token")

    user_id, session_id = row

    token = token_for_user(
        user_id=user_id, session_id=session_id, secret_key=request.app["secret_key"]
    )

    return json_response(
        data={
            "user_id": user_id,
            "access_token": token,
        }
    )


@routes.post("/auth/register")
@docs(tags=["auth"], summary="register a new user", description="")
@json_schema(
    Schema.from_dict(
        {"username": fields.Str(), "password": fields.Str(), "email": fields.Str()},
        name="RegisterSchema",
    )
)
async def register(request):
    data = request["json"]
    user_id = await request.app["user_service"].register_user(
        username=data["username"],
        password=data["password"],
        email=data["email"],
    )

    return json_response(data={"user_id": str(user_id)})


@routes.post("/auth/confirm_registration")
@docs(tags=["auth"], summary="confirm a pending registration", description="")
@json_schema(
    Schema.from_dict({"token": fields.Str()}, name="ConfirmRegistrationSchema")
)
async def confirm_registration(request):
    data = request["json"]
    try:
        await request.app["user_service"].confirm_registration(token=data["token"])
    except (PermissionError, InvalidCommand) as e:
        raise web.HTTPBadRequest(reason=str(e))

    return json_response(status=web.HTTPNoContent.status_code)


@routes.get("/profile")
@docs(tags=["auth"], summary="fetch user profile information", description="")
async def profile(request):
    user = await request.app["user_service"].get_user(
        user_id=request["user"]["user_id"]
    )

    serializer = UserSchema()

    return json_response(data=serializer.dump(user))


@routes.post("/profile/change_password")
@docs(tags=["auth"], summary="change password", description="")
@json_schema(
    Schema.from_dict(
        {"new_password": fields.Str(), "old_password": fields.Str()},
        name="ChangePasswordSchema",
    )
)
async def change_password(request):
    data = request["json"]
    try:
        await request.app["user_service"].change_password(
            user_id=request["user"]["user_id"],
            new_password=data["new_password"],
            old_password=data["old_password"],
        )
    except InvalidPassword as e:
        raise web.HTTPBadRequest(reason=str(e))

    return json_response(status=web.HTTPNoContent.status_code)


@routes.post("/profile/change_email")
@docs(tags=["auth"], summary="change email", description="")
@json_schema(
    Schema.from_dict(
        {"email": fields.Email(), "password": fields.Str()}, name="ChangeEmailSchema"
    )
)
async def change_email(request):
    data = request["json"]
    try:
        await request.app["user_service"].request_email_change(
            user_id=request["user"]["user_id"],
            email=data["email"],
            password=data["password"],
        )
    except InvalidPassword as e:
        raise web.HTTPBadRequest(reason=str(e))

    return json_response(status=web.HTTPNoContent.status_code)


@routes.post("/auth/confirm_email_change")
@docs(tags=["auth"], summary="confirm a pending email change", description="")
@json_schema(Schema.from_dict({"token": fields.Str()}, name="ConfirmEmailChangeSchema"))
async def confirm_email_change(request):
    data = request["json"]
    await request.app["user_service"].confirm_email_change(token=data["token"])

    return json_response(status=web.HTTPNoContent.status_code)


@routes.post("/auth/recover_password")
@docs(tags=["auth"], summary="recover password", description="")
@json_schema(Schema.from_dict({"email": fields.Email()}, name="RecoverPasswordSchema"))
async def recover_password(request):
    data = request["json"]
    try:
        await request.app["user_service"].request_password_recovery(
            email=data["email"],
        )
    except InvalidPassword as e:
        raise web.HTTPBadRequest(reason=str(e))

    return json_response(status=web.HTTPNoContent.status_code)


@routes.post("/auth/confirm_password_recovery")
@docs(tags=["auth"], summary="confirm a pending password recovery", description="")
@json_schema(
    Schema.from_dict(
        {"token": fields.Str(), "new_password": fields.Str()},
        name="ConfirmPasswordRecoverySchema",
    )
)
async def confirm_password_recovery(request):
    data = request["json"]
    try:
        await request.app["user_service"].confirm_password_recovery(
            token=data["token"], new_password=data["new_password"]
        )
    except PermissionError as e:
        raise web.HTTPBadRequest(reason=str(e))

    return json_response(status=web.HTTPNoContent.status_code)


@routes.post("/auth/delete_session")
@docs(tags=["auth"], summary="delete a given user session", description="")
@json_schema(Schema.from_dict({"session_id": fields.Int()}, name="DeleteSessionSchema"))
async def delete_session(request):
    data = request["json"]
    await request.app["user_service"].delete_session(
        user_id=request["user"]["user_id"], session_id=data["session_id"]
    )

    return json_response(status=web.HTTPNoContent.status_code)


@routes.post("/auth/rename_session")
@docs(tags=["auth"], summary="rename a given user session", description="")
@json_schema(
    Schema.from_dict(
        {"session_id": fields.Int(), "name": fields.Str()}, name="RenameSessionSchema"
    )
)
async def rename_session(request):
    data = request["json"]
    await request.app["user_service"].rename_session(
        user_id=request["user"]["user_id"],
        session_id=data["session_id"],
        name=data["name"],
    )

    return json_response(status=web.HTTPNoContent.status_code)



from datetime import datetime

import asyncpg

from abrechnung.application import (
    Application,
    NotFoundError,
    check_group_permissions,
    InvalidCommand,
    create_group_log,
)
from abrechnung.domain.groups import (
    Group,
    GroupMember,
    GroupPreview,
    GroupInvite,
    GroupLog,
)


class GroupService(Application):
    async def create_group(
        self,
        *,
        user_id: int,
        name: str,
        description: str,
        currency_symbol: str,
        terms: str,
    ) -> int:
        async with self.db_pool.acquire() as conn:
            async with conn.transaction():
                group_id = await conn.fetchval(
                    "insert into grp (name, description, currency_symbol, terms) "
                    "values ($1, $2, $3, $4) returning id",
                    name,
                    description,
                    currency_symbol,
                    terms,
                )
                await conn.execute(
                    "insert into group_membership (user_id, group_id, is_owner, can_write, description) "
                    "values ($1, $2, $3, $4, $5)",
                    user_id,
                    group_id,
                    True,
                    True,
                    "group founder",
                )
                await create_group_log(
                    conn=conn, group_id=group_id, user_id=user_id, type="group-created"
                )
                await create_group_log(
                    conn=conn,
                    group_id=group_id,
                    user_id=user_id,
                    type="member-joined",
                    affected_user_id=user_id,
                )

                return group_id

    async def create_invite(
        self,
        *,
        user_id: int,
        group_id: int,
        description: str,
        single_use: bool,
        join_as_editor: bool,
        valid_until: datetime,
    ) -> int:
        async with self.db_pool.acquire() as conn:
            async with conn.transaction():
                await check_group_permissions(
                    conn=conn, group_id=group_id, user_id=user_id, can_write=True
                )
                await create_group_log(
                    conn=conn, group_id=group_id, user_id=user_id, type="invite-created"
                )
                return await conn.fetchval(
                    "insert into group_invite(group_id, description, created_by, valid_until, single_use, join_as_editor)"
                    " values ($1, $2, $3, $4, $5, $6) returning id",
                    group_id,
                    description,
                    user_id,
                    valid_until,
                    single_use,
                    join_as_editor,
                )

    async def delete_invite(
        self,
        *,
        user_id: int,
        group_id: int,
        invite_id: int,
    ):
        async with self.db_pool.acquire() as conn:
            async with conn.transaction():
                await check_group_permissions(
                    conn=conn, group_id=group_id, user_id=user_id, can_write=True
                )
                deleted_id = await conn.fetchval(
                    "delete from group_invite where id = $1 and group_id = $2 returning id",
                    invite_id,
                    group_id,
                )
                if not deleted_id:
                    raise NotFoundError(f"No invite with the given id exists")
                await create_group_log(
                    conn=conn, group_id=group_id, user_id=user_id, type="invite-deleted"
                )

    async def join_group(self, user_id: int, invite_token: str):
        async with self.db_pool.acquire() as conn:
            async with conn.transaction():
                invite = await conn.fetchrow(
                    "select id, group_id, created_by, single_use, join_as_editor from group_invite gi "
                    "where gi.token = $1",
                    invite_token,
                )
                if not invite:
                    raise PermissionError(f"Invalid invite token")

                await conn.execute(
                    "insert into group_membership (user_id, group_id, invited_by, can_write, is_owner) "
                    "values ($1, $2, $3, $4, false)",
                    user_id,
                    invite["group_id"],
                    invite["created_by"],
                    invite["join_as_editor"],
                )
                await create_group_log(
                    conn=conn,
                    group_id=invite["group_id"],
                    user_id=user_id,
                    type="member-joined",
                    affected_user_id=user_id,
                )

                if invite["single_use"]:
                    await conn.execute(
                        "delete from group_invite where id = $1", invite["id"]
                    )

    async def list_groups(self, user_id: int) -> list[Group]:
        async with self.db_pool.acquire() as conn:
            async with conn.transaction():
                cur = conn.cursor(
                    "select grp.id, grp.name, grp.description, grp.terms, grp.currency_symbol, grp.created_at, "
                    "grp.created_by "
                    "from grp "
                    "join group_membership gm on grp.id = gm.group_id where gm.user_id = $1",
                    user_id,
                )
                result = []
                async for group in cur:
                    result.append(
                        Group(
                            id=group["id"],
                            name=group["name"],
                            description=group["description"],
                            currency_symbol=group["currency_symbol"],
                            terms=group["terms"],
                            created_at=group["created_at"],
                            created_by=group["created_by"],
                        )
                    )

            return result

    async def get_group(self, *, user_id: int, group_id: int) -> Group:
        async with self.db_pool.acquire() as conn:
            await check_group_permissions(conn=conn, group_id=group_id, user_id=user_id)
            group = await conn.fetchrow(
                "select id, name, description, terms, currency_symbol, created_at, created_by "
                "from grp "
                "where grp.id = $1",
                group_id,
            )
            if not group:
                raise NotFoundError(f"Group with id {group_id} does not exist")

            return Group(
                id=group["id"],
                name=group["name"],
                description=group["description"],
                currency_symbol=group["currency_symbol"],
                terms=group["terms"],
                created_at=group["created_at"],
                created_by=group["created_by"],
            )

    async def update_group(
        self,
        *,
        user_id: int,
        group_id: int,
        name: str,
        description: str,
        currency_symbol: str,
        terms: str,
    ):
        async with self.db_pool.acquire() as conn:
            async with conn.transaction():
                await check_group_permissions(
                    conn=conn, group_id=group_id, user_id=user_id, is_owner=True
                )
                await conn.execute(
                    "update grp set name = $2, description = $3, currency_symbol = $4, terms = $5 "
                    "where grp.id = $1",
                    group_id,
                    name,
                    description,
                    currency_symbol,
                    terms,
                )
                await create_group_log(
                    conn=conn, group_id=group_id, user_id=user_id, type="group-updated"
                )

    async def update_member_permissions(
        self,
        *,
        user_id: int,
        group_id: int,
        member_id: int,
        can_write: bool,
        is_owner: bool,
    ):
        async with self.db_pool.acquire() as conn:
            async with conn.transaction():
                if user_id == member_id:
                    raise InvalidCommand(
                        f"group members cannot modify their own privileges"
                    )

                # not possible to have an owner without can_write
                can_write = can_write if not is_owner else True

                user_can_write, user_is_owner = await check_group_permissions(
                    conn=conn, group_id=group_id, user_id=user_id, can_write=True
                )
                membership = await conn.fetchrow(
                    "select is_owner, can_write from group_membership where group_id = $1 and user_id = $2",
                    group_id,
                    member_id,
                )
                if membership is None:
                    raise NotFoundError(f"member with id {member_id} does not exist")

                if (
                    membership["is_owner"] == is_owner
                    and membership["can_write"] == can_write
                ):  # no changes
                    return

                if is_owner and not user_is_owner:
                    raise PermissionError(
                        f"group members cannot promote others to owner without being an owner"
                    )

                if is_owner:
                    await create_group_log(
                        conn=conn,
                        group_id=group_id,
                        user_id=user_id,
                        type="owner-granted",
                        affected_user_id=member_id,
                    )
                elif can_write:
                    if membership["is_owner"]:
                        await create_group_log(
                            conn=conn,
                            group_id=group_id,
                            user_id=user_id,
                            type="owner-revoked",
                            affected_user_id=member_id,
                        )
                    else:
                        await create_group_log(
                            conn=conn,
                            group_id=group_id,
                            user_id=user_id,
                            type="write-granted",
                            affected_user_id=member_id,
                        )
                else:
                    if membership["is_owner"]:
                        await create_group_log(
                            conn=conn,
                            group_id=group_id,
                            user_id=user_id,
                            type="owner-revoked",
                            affected_user_id=member_id,
                        )
                    await create_group_log(
                        conn=conn,
                        group_id=group_id,
                        user_id=user_id,
                        type="write-revoked",
                        affected_user_id=member_id,
                    )

                await conn.execute(
                    "update group_membership gm set can_write = $3, is_owner = $4 "
                    "where gm.user_id = $1 and gm.group_id = $2",
                    member_id,
                    group_id,
                    can_write,
                    is_owner,
                )

    async def delete_group(self, *, user_id: int, group_id: int):
        async with self.db_pool.acquire() as conn:
            async with conn.transaction():
                await check_group_permissions(
                    conn=conn, group_id=group_id, user_id=user_id, is_owner=True
                )

                n_members = await conn.fetchval(
                    "select count(user_id) from group_membership gm where gm.group_id = $1",
                    group_id,
                )
                if n_members != 1:
                    raise PermissionError(
                        f"Can only delete a group when you are the last member"
                    )

                await conn.execute("delete from grp where id = $1", group_id)

    async def leave_group(self, *, user_id: int, group_id: int):
        async with self.db_pool.acquire() as conn:
            async with conn.transaction():
                await check_group_permissions(
                    conn=conn, group_id=group_id, user_id=user_id
                )

                n_members = await conn.fetchval(
                    "select count(user_id) from group_membership gm where gm.group_id = $1",
                    group_id,
                )
                if (
                    n_members == 1
                ):  # our user is the last member -> delete the group, membership will be cascaded
                    await conn.execute("delete from grp where id = $1", group_id)
                else:
                    await conn.execute(
                        "delete from group_membership gm where gm.group_id = $1 and gm.user_id = $2",
                        group_id,
                        user_id,
                    )

    async def preview_group(self, invite_token: str) -> GroupPreview:
        async with self.db_pool.acquire() as conn:
            group = await conn.fetchrow(
                "select grp.id as group_id, "
                "grp.name, grp.description, grp.terms, grp.currency_symbol, grp.created_at, "
                "inv.description as invite_description, inv.valid_until as invite_valid_until, "
                "inv.single_use as invite_single_use "
                "from grp "
                "join group_invite inv on grp.id = inv.group_id "
                "where inv.token = $1",
                invite_token,
            )
            if not group:
                raise PermissionError(f"invalid invite token to preview group")

            return GroupPreview(
                id=group["group_id"],
                name=group["name"],
                description=group["description"],
                terms=group["terms"],
                currency_symbol=group["currency_symbol"],
                created_at=group["created_at"],
                invite_description=group["invite_description"],
                invite_valid_until=group["invite_valid_until"],
                invite_single_use=group["invite_single_use"],
            )

    async def list_invites(self, *, user_id: int, group_id: int) -> list[GroupInvite]:
        async with self.db_pool.acquire() as conn:
            async with conn.transaction():
                await check_group_permissions(
                    conn=conn, group_id=group_id, user_id=user_id
                )
                cur = conn.cursor(
                    "select id, case when created_by = $1 then token else null end as token, description, created_by, "
                    "valid_until, single_use, join_as_editor "
                    "from group_invite gi "
                    "where gi.group_id = $2",
                    user_id,
                    group_id,
                )
                result = []
                async for invite in cur:
                    result.append(
                        GroupInvite(
                            id=invite["id"],
                            token=invite["token"],
                            created_by=invite["created_by"],
                            valid_until=invite["valid_until"],
                            single_use=invite["single_use"],
                            description=invite["description"],
                            join_as_editor=invite["join_as_editor"],
                        )
                    )
                return result

    async def list_members(self, *, user_id: int, group_id: int) -> list[GroupMember]:
        async with self.db_pool.acquire() as conn:
            async with conn.transaction():
                await check_group_permissions(
                    conn=conn, group_id=group_id, user_id=user_id
                )
                cur = conn.cursor(
                    "select usr.id, usr.username, gm.is_owner, gm.can_write, gm.description, "
                    "gm.invited_by, gm.joined_at "
                    "from usr "
                    "join group_membership gm on gm.user_id = usr.id "
                    "where gm.group_id = $1",
                    group_id,
                )
                result = []
                async for group in cur:
                    result.append(
                        GroupMember(
                            user_id=group["id"],
                            username=group["username"],
                            is_owner=group["is_owner"],
                            can_write=group["can_write"],
                            invited_by=group["invited_by"],
                            joined_at=group["joined_at"],
                            description=group["description"],
                        )
                    )
                return result

    async def list_log(self, *, user_id: int, group_id: int) -> list[GroupLog]:
        async with self.db_pool.acquire() as conn:
            async with conn.transaction():
                await check_group_permissions(
                    conn=conn, group_id=group_id, user_id=user_id
                )
                cur = conn.cursor(
                    "select id, user_id, logged_at, type, message, affected "
                    "from group_log "
                    "where group_id = $1",
                    group_id,
                )

                result = []
                async for log in cur:
                    result.append(
                        GroupLog(
                            id=log["id"],
                            user_id=log["user_id"],
                            logged_at=log["logged_at"],
                            type=log["type"],
                            message=log["message"],
                            affected=log["affected"],
                        )
                    )
                return result

    async def send_group_message(self, *, user_id: int, group_id: int, message: str):
        async with self.db_pool.acquire() as conn:
            async with conn.transaction():
                await check_group_permissions(
                    conn=conn, group_id=group_id, user_id=user_id, can_write=True
                )
                await conn.execute(
                    "insert into group_log (group_id, user_id, type, message) "
                    "values ($1, $2, 'text-message', $3)",
                    group_id,
                    user_id,
                    message,
                )



from datetime import timedelta, datetime, timezone

from abrechnung.http.auth import token_for_user
from tests.http_tests import HTTPAPITest


class GroupAPITest(HTTPAPITest):
    async def _fetch_group(self, group_id: int, expected_status: int = 200) -> dict:
        resp = await self._get(f"/api/v1/groups/{group_id}")
        self.assertEqual(expected_status, resp.status)
        if expected_status >= 400:
            return {}
        ret_data = await resp.json()
        self.assertEqual(group_id, ret_data["id"])
        return ret_data

    async def _fetch_account(self, account_id: int, expected_status: int = 200) -> dict:
        resp = await self._get(f"/api/v1/accounts/{account_id}")
        self.assertEqual(expected_status, resp.status)
        if expected_status >= 400:
            return {}
        ret_data = await resp.json()
        self.assertEqual(account_id, ret_data["id"])
        return ret_data

    async def _fetch_members(self, group_id: int, expected_status: int = 200) -> dict:
        resp = await self._get(f"/api/v1/groups/{group_id}/members")
        self.assertEqual(expected_status, resp.status)
        if expected_status >= 400:
            return {}
        ret_data = await resp.json()
        self.assertTrue(isinstance(ret_data, list))
        return ret_data

    async def test_create_group(self):
        resp = await self._post(
            "/api/v1/groups",
            json={
                "name": "name",
                "description": "description",
                "currency_symbol": "€",
                "terms": "terms",
            },
        )
        self.assertEqual(200, resp.status)
        ret_data = await resp.json()
        self.assertIsNotNone(ret_data["group_id"])
        group_id = ret_data["group_id"]

        group = await self._fetch_group(group_id)
        self.assertEqual("name", group["name"])

        await self._fetch_group("asdf1234", 404)
        await self._fetch_group(13333, 404)

        resp = await self._post(
            f"/api/v1/groups/{group_id}",
            json={
                "name": "name2",
                "description": "description2",
                "currency_symbol": "$",
                "terms": "new terms",
            },
        )
        self.assertEqual(204, resp.status)

        group = await self._fetch_group(group_id)
        self.assertEqual("name2", group["name"])
        self.assertEqual("description2", group["description"])
        self.assertEqual("$", group["currency_symbol"])
        self.assertEqual("new terms", group["terms"])

    async def test_delete_group(self):
        user2_id, user2_password = await self._create_test_user(
            "user2", "user2@test.com"
        )
        _, session_id, _ = await self.user_service.login_user(
            "user2", password=user2_password, session_name="foobar"
        )
        user2_token = token_for_user(user2_id, session_id, self.secret_key)
        group_id = await self.group_service.create_group(
            user_id=self.test_user_id,
            name="foobar",
            description="foobar",
            currency_symbol="€",
            terms="foo",
        )
        async with self.db_pool.acquire() as conn:
            await conn.execute(
                "insert into group_membership (user_id, group_id, is_owner, can_write) "
                "values ($1, $2, true, true)",
                user2_id,
                group_id,
            )

        account1_id = await self.account_service.create_account(
            user_id=self.test_user_id,
            group_id=group_id,
            type="personal",
            name="account1",
            description="",
        )
        account2_id = await self.account_service.create_account(
            user_id=self.test_user_id,
            group_id=group_id,
            type="personal",
            name="account2",
            description="",
        )

        transaction_id = await self.transaction_service.create_transaction(
            user_id=self.test_user_id,
            group_id=group_id,
            type="purchase",
            description="asdf",
            billed_at=datetime.now(tz=timezone.utc),
            currency_symbol="€",
            currency_conversion_rate=1.0,
            value=20.0,
        )
        await self.transaction_service.add_or_change_debitor_share(
            user_id=self.test_user_id,
            transaction_id=transaction_id,
            account_id=account1_id,
            value=1.0,
        )
        await self.transaction_service.add_or_change_creditor_share(
            user_id=self.test_user_id,
            transaction_id=transaction_id,
            account_id=account2_id,
            value=1.0,
        )

        resp = await self._delete(f"/api/v1/groups/{group_id}")
        self.assertEqual(403, resp.status)

        resp = await self._post(f"/api/v1/groups/{group_id}/leave")
        self.assertEqual(204, resp.status)

        await self._fetch_group(group_id, expected_status=404)

        resp = await self.client.delete(
            f"/api/v1/groups/{group_id}",
            headers={"Authorization": f"Bearer {user2_token}"},
        )
        self.assertEqual(204, resp.status)

        async with self.db_pool.acquire() as conn:
            gid = await conn.fetchval("select id from grp where grp.id = $1", group_id)
            self.assertIsNone(gid)

    async def test_list_groups(self):
        resp = await self._get(f"/api/v1/groups")
        self.assertEqual(200, resp.status)
        ret_data = await resp.json()
        self.assertEqual(0, len(ret_data))

        group1_id = await self.group_service.create_group(
            user_id=self.test_user_id,
            name="name1",
            description="description",
            currency_symbol="€",
            terms="terms",
        )
        group2_id = await self.group_service.create_group(
            user_id=self.test_user_id,
            name="name2",
            description="description",
            currency_symbol="€",
            terms="terms",
        )
        resp = await self._get(f"/api/v1/groups")
        self.assertEqual(200, resp.status)
        ret_data = await resp.json()
        self.assertEqual(2, len(ret_data))
        self.assertEqual({group1_id, group2_id}, set([e["id"] for e in ret_data]))

    async def test_create_account(self):
        group_id = await self.group_service.create_group(
            user_id=self.test_user_id,
            name="name",
            description="description",
            currency_symbol="€",
            terms="terms",
        )
        resp = await self._post(
            f"/api/v1/groups/{group_id}/accounts",
            json={
                "name": "name",
                "description": "description",
                "type": "personal",
            },
        )
        self.assertEqual(200, resp.status)
        ret_data = await resp.json()
        self.assertIsNotNone(ret_data["id"])

    async def test_update_account(self):
        group_id = await self.group_service.create_group(
            user_id=self.test_user_id,
            name="name",
            description="description",
            currency_symbol="€",
            terms="terms",
        )
        account_id = await self.account_service.create_account(
            user_id=self.test_user_id,
            group_id=group_id,
            type="personal",
            name="name",
            description="description",
        )
        resp = await self._post(
            f"/api/v1/accounts/{account_id}",
            json={
                "name": "new_name",
                "description": "description",
            },
        )
        self.assertEqual(200, resp.status)

        a = await self._fetch_account(account_id)
        self.assertEqual("new_name", a["name"])
        self.assertEqual("description", a["description"])

        resp = await self._post(
            f"/api/v1/accounts/{account_id}",
            json={
                "name": "new_name2",
                "description": "description1",
            },
        )
        self.assertEqual(200, resp.status)
        a = await self._fetch_account(account_id)
        self.assertEqual("new_name2", a["name"])
        self.assertEqual("description1", a["description"])

    async def test_list_accounts(self):
        group_id = await self.group_service.create_group(
            user_id=self.test_user_id,
            name="name",
            description="description",
            currency_symbol="€",
            terms="terms",
        )
        account1_id = await self.account_service.create_account(
            user_id=self.test_user_id,
            group_id=group_id,
            type="personal",
            name="account1",
            description="description",
        )
        account2_id = await self.account_service.create_account(
            user_id=self.test_user_id,
            group_id=group_id,
            type="personal",
            name="account2",
            description="description",
        )
        resp = await self._get(f"/api/v1/groups/{group_id}/accounts")
        self.assertEqual(200, resp.status)
        ret_data = await resp.json()
        self.assertEqual(2, len(ret_data))
        self.assertEqual({account1_id, account2_id}, set([e["id"] for e in ret_data]))

    async def test_members(self):
        group_id = await self.group_service.create_group(
            user_id=self.test_user_id,
            name="name",
            description="description",
            currency_symbol="€",
            terms="terms",
        )
        members = await self._fetch_members(group_id)
        self.assertEqual(1, len(members))
        self.assertEqual({self.test_user_id}, set([m["user_id"] for m in members]))

        user2_id, _ = await self._create_test_user("user2", "user2@email.stuff")
        async with self.db_pool.acquire() as conn:
            await conn.execute(
                "insert into group_membership (user_id, group_id, invited_by) values ($1, $2, $3)",
                user2_id,
                group_id,
                self.test_user_id,
            )

        members = await self._fetch_members(group_id)
        self.assertEqual(2, len(members))
        self.assertEqual(
            {self.test_user_id, user2_id}, set([m["user_id"] for m in members])
        )

        resp = await self._post(
            f"/api/v1/groups/{group_id}/members",
            json={"user_id": user2_id, "can_write": True, "is_owner": False},
        )
        self.assertEqual(204, resp.status)

        members = await self._fetch_members(group_id)
        self.assertEqual(2, len(members))
        self.assertTrue(
            list(filter(lambda x: x["user_id"] == user2_id, members))[0]["can_write"]
        )
        self.assertFalse(
            list(filter(lambda x: x["user_id"] == user2_id, members))[0]["is_owner"]
        )

        resp = await self._post(
            f"/api/v1/groups/{group_id}/members",
            json={"user_id": user2_id, "can_write": True, "is_owner": True},
        )
        self.assertEqual(204, resp.status)

        members = await self._fetch_members(group_id)
        self.assertEqual(2, len(members))
        self.assertTrue(
            list(filter(lambda x: x["user_id"] == user2_id, members))[0]["can_write"]
        )
        self.assertTrue(
            list(filter(lambda x: x["user_id"] == user2_id, members))[0]["is_owner"]
        )

    async def test_get_account(self):
        group_id = await self.group_service.create_group(
            user_id=self.test_user_id,
            name="name",
            description="description",
            currency_symbol="€",
            terms="terms",
        )
        account_id = await self.account_service.create_account(
            user_id=self.test_user_id,
            group_id=group_id,
            type="personal",
            name="account1",
            description="description",
        )
        ret_data = await self._fetch_account(account_id)
        self.assertEqual("account1", ret_data["name"])

        resp = await self._get(f"/api/v1/accounts/asdf1234")
        self.assertEqual(404, resp.status)

        resp = await self._get(f"/api/v1/accounts/13232")
        self.assertEqual(404, resp.status)

    async def test_invites(self):
        group_id = await self.group_service.create_group(
            user_id=self.test_user_id,
            name="group1",
            description="description",
            currency_symbol="€",
            terms="terms",
        )
        resp = await self._post(
            f"/api/v1/groups/{group_id}/invites",
            json={
                "description": "description",
                "single_use": False,
                "join_as_editor": False,
                "valid_until": (datetime.now() + timedelta(hours=1)).isoformat(),
            },
        )
        self.assertEqual(204, resp.status)

        resp = await self._get(f"/api/v1/groups/{group_id}/invites")
        self.assertEqual(200, resp.status)
        invites = await resp.json()
        self.assertEqual(1, len(invites))

        # we hardcode assume the invite id is 0 to check if the id counting works as expected
        resp = await self._delete(
            f"/api/v1/groups/{group_id}/invites/{invites[0]['id']}"
        )
        self.assertEqual(204, resp.status)

        resp = await self._get(f"/api/v1/groups/{group_id}/invites")
        self.assertEqual(200, resp.status)
        invites = await resp.json()
        self.assertEqual(0, len(invites))

        # now check that we can actually preview the group as a different user
        resp = await self._post(
            f"/api/v1/groups/{group_id}/invites",
            json={
                "description": "description",
                "single_use": True,
                "join_as_editor": True,
                "valid_until": (datetime.now() + timedelta(hours=1)).isoformat(),
            },
        )
        self.assertEqual(204, resp.status)

        resp = await self._get(f"/api/v1/groups/{group_id}/invites")
        self.assertEqual(200, resp.status)
        invites = await resp.json()
        self.assertEqual(1, len(invites))
        invite_token = invites[0]["token"]
        self.assertIsNotNone(invite_token)

        user2_id, password = await self._create_test_user("user", "email2@email.stuff")
        _, session_id, _ = await self.user_service.login_user(
            "user", password=password, session_name="session1"
        )
        jwt_token = token_for_user(
            user2_id, session_id=session_id, secret_key=self.secret_key
        )
        resp = await self.client.post(
            f"/api/v1/groups/preview",
            json={"invite_token": invite_token},
            headers={"Authorization": f"Bearer {jwt_token}"},
        )
        self.assertEqual(200, resp.status)
        prev_group_data = await resp.json()
        self.assertEqual("group1", prev_group_data["name"])
        self.assertEqual(group_id, prev_group_data["id"])

        resp = await self.client.post(
            f"/api/v1/groups/join",
            json={"invite_token": invite_token},
            headers={"Authorization": f"Bearer {jwt_token}"},
        )
        self.assertEqual(204, resp.status)

        resp = await self.client.get(
            f"/api/v1/groups",
            headers={"Authorization": f"Bearer {jwt_token}"},
        )
        self.assertEqual(200, resp.status)
        groups = await resp.json()
        self.assertEqual(1, len(groups))

        resp = await self._get(f"/api/v1/groups/{group_id}/invites")
        self.assertEqual(200, resp.status)
        invites = await resp.json()
        self.assertEqual(0, len(invites))

        resp = await self._get(f"/api/v1/groups/{group_id}/logs")
        self.assertEqual(200, resp.status)
        logs = await resp.json()
        self.assertEqual(6, len(logs))
        self.assertEqual(
            {"group-created", "member-joined", "invite-created", "invite-deleted"},
            set([log["type"] for log in logs]),
        )

    async def test_group_log(self):
        group_id = await self.group_service.create_group(
            user_id=self.test_user_id,
            name="group1",
            description="description",
            currency_symbol="€",
            terms="terms",
        )

        resp = await self._post(
            f"/api/v1/groups/{group_id}/send_message", json={"message": "test message"}
        )
        self.assertEqual(204, resp.status)

        resp = await self._get(f"/api/v1/groups/{group_id}/logs")
        self.assertEqual(200, resp.status)
        logs = await resp.json()
        self.assertEqual(3, len(logs))
        self.assertEqual(
            {"group-created", "member-joined", "text-message"},
            set([log["type"] for log in logs]),
        )



from datetime import datetime

from aiohttp import web
from aiohttp.abc import Request
from aiohttp.web_request import FileField
from marshmallow import Schema, fields

from abrechnung.http.openapi import docs, json_schema
from abrechnung.http.serializers import TransactionSchema
from abrechnung.http.utils import json_response

routes = web.RouteTableDef()


async def _transaction_response(request, transaction_id: int) -> web.Response:
    transaction = await request.app["transaction_service"].get_transaction(
        user_id=request["user"]["user_id"], transaction_id=transaction_id
    )

    serializer = TransactionSchema()
    return json_response(data=serializer.dump(transaction))


@routes.get(r"/groups/{group_id:\d+}/transactions")
@docs(
    tags=["transactions"],
    summary="list all transactions in a group",
    responses={200: {"schema": TransactionSchema(many=True), "description": ""}},
    description="",
)
async def list_transactions(request):
    group_id: int = int(request.match_info["group_id"])
    min_last_changed = request.query.get("min_last_changed")

    if min_last_changed:
        try:
            min_last_changed = datetime.fromisoformat(min_last_changed)
        except ValueError:
            raise web.HTTPBadRequest(
                reason="Invalid query param 'min_last_changed', must be a valid ISO timestamp."
            )

    forced_transaction_ids = request.query.get("transaction_ids")
    if forced_transaction_ids:
        try:
            forced_transaction_ids = [int(x) for x in forced_transaction_ids.split(",")]
        except ValueError:
            raise web.HTTPBadRequest(
                reason="Invalid query param 'transaction_ids', must be a comma separated list of integers"
            )

    transactions = await request.app["transaction_service"].list_transactions(
        user_id=request["user"]["user_id"],
        group_id=group_id,
        min_last_changed=min_last_changed,
        additional_transactions=forced_transaction_ids,
    )

    serializer = TransactionSchema()
    return json_response(data=serializer.dump(transactions, many=True))


@routes.post(r"/groups/{group_id:\d+}/transactions")
@docs(
    tags=["transactions"],
    summary="create a new transaction",
    responses={200: {"schema": TransactionSchema, "description": ""}},
    description="",
)
@json_schema(
    Schema.from_dict(
        {
            "description": fields.Str(),
            "type": fields.Str(),
            "value": fields.Number(),
            "currency_symbol": fields.Str(),
            "billed_at": fields.Date(),
            "currency_conversion_rate": fields.Number(),
        },
        name="CreateTransactionSchema",
    )
)
async def create_transaction(request: Request):
    data = request["json"]
    group_id: int = int(request.match_info["group_id"])

    transaction_id = await request.app["transaction_service"].create_transaction(
        user_id=request["user"]["user_id"],
        group_id=group_id,
        description=data["description"],
        type=data["type"],
        currency_symbol=data["currency_symbol"],
        currency_conversion_rate=float(data["currency_conversion_rate"]),
        billed_at=data["billed_at"],
        value=float(data["value"]),
    )

    return await _transaction_response(request, transaction_id)


@routes.get(r"/transactions/{transaction_id:\d+}")
@docs(
    tags=["transactions"],
    summary="get transaction details",
    responses={200: {"schema": TransactionSchema, "description": ""}},
    description="",
)
async def get_transaction(request: Request):
    transaction = await request.app["transaction_service"].get_transaction(
        user_id=request["user"]["user_id"],
        transaction_id=int(request.match_info["transaction_id"]),
    )

    serializer = TransactionSchema()
    return json_response(data=serializer.dump(transaction))


@routes.post(r"/transactions/{transaction_id:\d+}")
@docs(
    tags=["transactions"],
    summary="update transaction details",
    responses={200: {"schema": TransactionSchema, "description": ""}},
    description="",
)
@json_schema(
    Schema.from_dict(
        {
            "description": fields.Str(),
            "value": fields.Number(),
            "currency_symbol": fields.Str(),
            "billed_at": fields.Date(),
            "currency_conversion_rate": fields.Number(),
        },
        name="UpdateTransactionSchema",
    )
)
async def update_transaction(request: Request):
    data = request["json"]
    transaction_id = int(request.match_info["transaction_id"])

    await request.app["transaction_service"].update_transaction(
        user_id=request["user"]["user_id"],
        transaction_id=transaction_id,
        value=data["value"],
        description=data["description"],
        currency_symbol=data["currency_symbol"],
        currency_conversion_rate=data["currency_conversion_rate"],
        billed_at=data["billed_at"],
    )

    return await _transaction_response(request, transaction_id)


@routes.post(r"/transactions/{transaction_id:\d+}/commit")
@docs(
    tags=["transactions"],
    summary="commit currently pending transaction changes",
    responses={200: {"schema": TransactionSchema, "description": ""}},
    description="",
)
async def commit_transaction(request: Request):
    transaction_id = int(request.match_info["transaction_id"])
    await request.app["transaction_service"].commit_transaction(
        user_id=request["user"]["user_id"],
        transaction_id=transaction_id,
    )

    return await _transaction_response(request, transaction_id)


@routes.delete(r"/transactions/{transaction_id:\d+}")
@docs(
    tags=["transactions"],
    summary="delete a transaction",
    responses={200: {"schema": TransactionSchema, "description": ""}},
    description="",
)
async def delete_transaction(request: Request):
    transaction_id = int(request.match_info["transaction_id"])
    await request.app["transaction_service"].delete_transaction(
        user_id=request["user"]["user_id"],
        transaction_id=transaction_id,
    )

    return await _transaction_response(request, transaction_id)


@routes.post(r"/transactions/{transaction_id:\d+}/new_change")
@docs(
    tags=["transactions"],
    summary="create a new pending transaction revision",
    responses={200: {"schema": TransactionSchema, "description": ""}},
    description="",
)
async def create_transaction_change(request: Request):
    transaction_id = int(request.match_info["transaction_id"])
    await request.app["transaction_service"].create_transaction_change(
        user_id=request["user"]["user_id"],
        transaction_id=int(request.match_info["transaction_id"]),
    )

    return await _transaction_response(request, transaction_id)


@routes.post(r"/transactions/{transaction_id:\d+}/discard")
@docs(
    tags=["transactions"],
    summary="discard currently pending transaction changes",
    responses={200: {"schema": TransactionSchema, "description": ""}},
    description="",
)
async def discard_transaction_change(request: Request):
    transaction_id = int(request.match_info["transaction_id"])
    await request.app["transaction_service"].discard_transaction_changes(
        user_id=request["user"]["user_id"],
        transaction_id=transaction_id,
    )

    return await _transaction_response(request, transaction_id)


class ChangeShareResponse(Schema):
    account_id = fields.Int()
    value = fields.Number()


class DeleteShareResponse(Schema):
    account_id = fields.Int()


@routes.post(r"/transactions/{transaction_id:\d+}/creditor_shares")
@docs(
    tags=["transactions"],
    summary="add or update a creditor share",
    responses={200: {"schema": TransactionSchema, "description": ""}},
    description="",
)
@json_schema(ChangeShareResponse)
async def add_or_change_creditor_share(request: Request):
    data = request["json"]
    transaction_id = int(request.match_info["transaction_id"])
    await request.app["transaction_service"].add_or_change_creditor_share(
        user_id=request["user"]["user_id"],
        transaction_id=transaction_id,
        account_id=data["account_id"],
        value=float(data["value"]),
    )

    return await _transaction_response(request, transaction_id)


@routes.post(r"/transactions/{transaction_id:\d+}/creditor_shares/switch")
@docs(
    tags=["transactions"],
    summary="switch the current creditor share",
    description="For transfer type transactions only one creditor share is allowed, "
    "this allows switching the currently selected creditor share with a new one. "
    "This is essentially a shortcut to deleting the current share and adding a new one.",
    responses={200: {"schema": TransactionSchema, "description": ""}},
)
@json_schema(ChangeShareResponse)
async def switch_creditor_share(request: Request):
    data = request["json"]
    transaction_id = int(request.match_info["transaction_id"])
    await request.app["transaction_service"].switch_creditor_share(
        user_id=request["user"]["user_id"],
        transaction_id=transaction_id,
        account_id=data["account_id"],
        value=float(data["value"]),
    )

    return await _transaction_response(request, transaction_id)


@routes.delete(r"/transactions/{transaction_id:\d+}/creditor_shares")
@docs(
    tags=["transactions"],
    summary="delete a creditor share",
    responses={200: {"schema": TransactionSchema, "description": ""}},
    description="",
)
@json_schema(DeleteShareResponse)
async def delete_creditor_share(request: Request):
    data = request["json"]
    transaction_id = int(request.match_info["transaction_id"])
    await request.app["transaction_service"].delete_creditor_share(
        user_id=request["user"]["user_id"],
        transaction_id=transaction_id,
        account_id=data["account_id"],
    )

    return await _transaction_response(request, transaction_id)


@routes.post(r"/transactions/{transaction_id:\d+}/debitor_shares")
@docs(
    tags=["transactions"],
    summary="add or change a debitor share",
    responses={200: {"schema": TransactionSchema, "description": ""}},
    description="",
)
@json_schema(ChangeShareResponse)
async def add_or_change_debitor_share(request: Request):
    data = request["json"]
    transaction_id = int(request.match_info["transaction_id"])
    await request.app["transaction_service"].add_or_change_debitor_share(
        user_id=request["user"]["user_id"],
        transaction_id=transaction_id,
        account_id=data["account_id"],
        value=float(data["value"]),
    )

    return await _transaction_response(request, transaction_id)


@routes.post(r"/transactions/{transaction_id:\d+}/debitor_shares/switch")
@docs(
    tags=["transactions"],
    summary="switch the current debitor share",
    description="For transfer and purchase type transactions only one debitor share is allowed, "
    "this allows switching the currently selected debitor share with a new one. "
    "This is essentially a shortcut to deleting the current share and adding a new one.",
    responses={200: {"schema": TransactionSchema, "description": ""}},
)
@json_schema(ChangeShareResponse)
async def switch_debitor_share(request: Request):
    data = request["json"]
    transaction_id = int(request.match_info["transaction_id"])
    await request.app["transaction_service"].switch_debitor_share(
        user_id=request["user"]["user_id"],
        transaction_id=transaction_id,
        account_id=data["account_id"],
        value=float(data["value"]),
    )

    return await _transaction_response(request, transaction_id)


@routes.delete(r"/transactions/{transaction_id:\d+}/debitor_shares")
@docs(
    tags=["transactions"],
    summary="delete a debitor share",
    responses={200: {"schema": TransactionSchema, "description": ""}},
    description="",
)
@json_schema(DeleteShareResponse)
async def delete_debitor_share(request: Request):
    data = request["json"]
    transaction_id = int(request.match_info["transaction_id"])
    await request.app["transaction_service"].delete_debitor_share(
        user_id=request["user"]["user_id"],
        transaction_id=transaction_id,
        account_id=data["account_id"],
    )

    return await _transaction_response(request, transaction_id)


class PositionCreateResponse(Schema):
    transaction = fields.Nested(TransactionSchema)
    item_id = fields.Int()


@routes.post(r"/transactions/{transaction_id:\d+}/purchase_items")
@docs(
    tags=["transactions"],
    summary="create a transaction position",
    description="",
    responses={
        200: {
            "schema": PositionCreateResponse,
            "description": "",
        }
    },
)
@json_schema(
    Schema.from_dict(
        {
            "name": fields.Str(),
            "price": fields.Number(),
            "communist_shares": fields.Number(),
            "usages": fields.Dict(
                keys=fields.Int(), values=fields.Number(), required=False, missing=None
            ),
        },
        name="CreatePositionSchema",
    )
)
async def create_purchase_item(request: Request):
    transaction_id = int(request.match_info["transaction_id"])
    data = request["json"]
    item_id = await request.app["transaction_service"].create_purchase_item(
        user_id=request["user"]["user_id"],
        transaction_id=transaction_id,
        name=data["name"],
        price=data["price"],
        communist_shares=data["communist_shares"],
        usages=data["usages"],
    )
    transaction = await request.app["transaction_service"].get_transaction(
        user_id=request["user"]["user_id"], transaction_id=transaction_id
    )

    serializer = TransactionSchema()

    return json_response(
        data={"transaction": serializer.dump(transaction), "item_id": item_id}
    )


@routes.post(r"/purchase_items/{item_id:\d+}")
@docs(
    tags=["transactions"],
    summary="update a transaction position",
    description="",
    responses={200: {"schema": TransactionSchema, "description": ""}},
)
@json_schema(
    Schema.from_dict(
        {
            "name": fields.Str(),
            "price": fields.Number(),
            "communist_shares": fields.Number(),
        },
        name="UpdatePositionSchema",
    )
)
async def update_purchase_item(request: Request):
    data = request["json"]
    transaction_id, revision_id = await request.app[
        "transaction_service"
    ].update_purchase_item(
        user_id=request["user"]["user_id"],
        item_id=int(request.match_info["item_id"]),
        name=data["name"],
        price=data["price"],
        communist_shares=data["communist_shares"],
    )

    return await _transaction_response(request, transaction_id)


@routes.post(r"/purchase_items/{item_id:\d+}/shares")
@docs(
    tags=["transactions"],
    summary="add or change a transaction position share",
    description="",
    responses={200: {"schema": TransactionSchema, "description": ""}},
)
@json_schema(
    Schema.from_dict(
        {
            "account_id": fields.Int(),
            "share_amount": fields.Number(),
        },
        name="ChangeShareSchema",
    )
)
async def add_or_change_item_share(request: Request):
    data = request["json"]
    transaction_id, revision_id = await request.app[
        "transaction_service"
    ].add_or_change_item_share(
        user_id=request["user"]["user_id"],
        item_id=int(request.match_info["item_id"]),
        account_id=data["account_id"],
        share_amount=data["share_amount"],
    )

    return await _transaction_response(request, transaction_id)


@routes.delete(r"/purchase_items/{item_id:\d+}/shares")
@docs(
    tags=["transactions"],
    summary="delete a position share",
    description="",
    responses={200: {"schema": TransactionSchema, "description": ""}},
)
@json_schema(DeleteShareResponse)
async def delete_item_share(request: Request):
    data = request["json"]
    transaction_id, revision_id = await request.app[
        "transaction_service"
    ].delete_item_share(
        user_id=request["user"]["user_id"],
        item_id=int(request.match_info["item_id"]),
        account_id=data["account_id"],
    )

    return await _transaction_response(request, transaction_id)


@routes.delete(r"/purchase_items/{item_id:\d+}")
@docs(
    tags=["transactions"],
    summary="delete a transaction position",
    description="",
    responses={200: {"schema": TransactionSchema, "description": ""}},
)
async def delete_purchase_item(request: Request):
    transaction_id, revision_id = await request.app[
        "transaction_service"
    ].delete_purchase_item(
        user_id=request["user"]["user_id"], item_id=int(request.match_info["item_id"])
    )

    return await _transaction_response(request, transaction_id)


@routes.post(r"/transactions/{transaction_id:\d+}/files")
@docs(
    tags=["transactions"],
    summary="upload a file as a transaction attachment",
    description="",
    responses={200: {"schema": TransactionSchema, "description": ""}},
)
async def upload_file(request: Request):
    data = await request.post()

    if "file" not in data or not isinstance(data["file"], FileField):
        raise web.HTTPBadRequest(reason=f"File data required")
    if "filename" not in data or not isinstance(data["filename"], str):
        raise web.HTTPBadRequest(reason=f"File Name required")

    transaction_id = int(request.match_info["transaction_id"])
    file_field: FileField = data["file"]
    # filename contains the name of the file in string format.
    filename = data["filename"]
    input_file = file_field.file
    mime_type = file_field.content_type
    try:
        content = input_file.read()
    except Exception as e:
        raise web.HTTPBadRequest(reason=f"Cannot read uploaded file: {e}")

    await request.app["transaction_service"].upload_file(
        user_id=request["user"]["user_id"],
        transaction_id=transaction_id,
        filename=filename,
        mime_type=mime_type,
        content=content,
    )

    return await _transaction_response(request, transaction_id)


@routes.delete(r"/files/{file_id:\d+}")
@docs(
    tags=["transactions"],
    summary="delete a transaction attachment",
    description="",
    responses={200: {"schema": TransactionSchema, "description": ""}},
)
async def delete_file(request: Request):
    transaction_id, revision_id = await request.app["transaction_service"].delete_file(
        user_id=request["user"]["user_id"],
        file_id=int(request.match_info["file_id"]),
    )

    return await _transaction_response(request, transaction_id)


@routes.get(r"/files/{file_id:\d+}/{blob_id:\d+}")
@docs(
    tags=["transactions"],
    summary="fetch the (binary) contents of a transaction attachment",
    description="",
)
async def get_file_contents(request: Request):
    file_id = int(request.match_info["file_id"])
    blob_id = int(request.match_info["blob_id"])
    mime_type, content = await request.app["transaction_service"].read_file_contents(
        user_id=request["user"]["user_id"],
        file_id=file_id,
        blob_id=blob_id,
    )

    return web.Response(body=content, content_type=mime_type)



def docs(**kwargs):
    """
    Annotate the decorated view function with the specified Swagger
    attributes.

    Usage:

    .. code-block:: python

        from aiohttp import web

        @docs(tags=['my_tag'],
              summary='Test method summary',
              description='Test method description',
              parameters=[{
                      'in': 'header',
                      'name': 'X-Request-ID',
                      'schema': {'type': 'string', 'format': 'uuid'},
                      'required': 'true'
                  }]
              )
        async def index(request):
            return web.json_response({'msg': 'done', 'data': {}})

    """

    def wrapper(func):
        if not kwargs.get("produces"):
            kwargs["produces"] = ["application/json"]
        if not hasattr(func, "__apispec__"):
            func.__apispec__ = {"schemas": [], "responses": {}, "parameters": []}
            func.__schemas__ = []
        extra_parameters = kwargs.pop("parameters", [])
        extra_responses = kwargs.pop("responses", {})
        func.__apispec__["parameters"].extend(extra_parameters)
        func.__apispec__["responses"].update(extra_responses)
        func.__apispec__.update(kwargs)
        return func

    return wrapper



from dataclasses import dataclass, field
from datetime import datetime, date
from enum import Enum
from typing import Optional


class TransactionType(Enum):
    mimo = "mimo"
    purchase = "purchase"
    transfer = "transfer"


@dataclass
class TransactionPosition:
    id: int

    name: str
    price: float
    communist_shares: float
    deleted: bool

    # usages map account IDs to portions of the item share pool
    usages: dict[int, float]


@dataclass
class TransactionDetails:
    description: str
    value: float
    currency_symbol: str
    currency_conversion_rate: float
    billed_at: date
    deleted: bool

    changed_by: int
    committed_at: Optional[datetime]

    # creditor and debitor shares map account IDs to portions of the communist share pool
    creditor_shares: dict[int, float]
    debitor_shares: dict[int, float]


@dataclass
class FileAttachment:
    id: int
    filename: str
    blob_id: Optional[int]
    mime_type: Optional[str]
    host_url: str
    deleted: bool


@dataclass
class Transaction:
    id: int
    group_id: int
    type: str
    is_wip: bool
    last_changed: datetime
    version: int
    committed_details: Optional[TransactionDetails]
    pending_details: Optional[TransactionDetails]

    committed_positions: Optional[list[TransactionPosition]] = field(default=None)
    pending_positions: Optional[list[TransactionPosition]] = field(default=None)

    committed_files: Optional[list[FileAttachment]] = field(default=None)
    pending_files: Optional[list[FileAttachment]] = field(default=None)



from dataclasses import dataclass
from datetime import datetime
from typing import Optional


@dataclass
class GroupMember:
    user_id: int
    username: str
    is_owner: bool
    can_write: bool
    description: str
    joined_at: datetime
    invited_by: Optional[int]


@dataclass
class GroupInvite:
    id: int
    created_by: int
    token: str
    single_use: bool
    join_as_editor: bool
    description: str
    valid_until: datetime


@dataclass
class Group:
    id: int
    name: str
    description: str
    currency_symbol: str
    terms: str
    created_at: datetime
    created_by: int


@dataclass
class GroupLog:
    id: int
    user_id: int
    logged_at: datetime
    type: str
    message: str
    affected: Optional[int]


@dataclass
class GroupPreview:
    id: int
    name: str
    description: str
    currency_symbol: str
    terms: str
    created_at: datetime
    invite_single_use: bool
    invite_valid_until: datetime
    invite_description: str



import logging
from typing import Callable, Optional

from aiohttp import web

logger = logging.getLogger(__name__)


def setup_aiohttp_apispec(
    app: web.Application,
    *,
    title: str = "API documentation",
    version: str = "0.0.1",
    url: str = "/api/docs/swagger.json",
    request_data_name: str = "data",
    error_callback=None,
    in_place: bool = False,
    prefix: str = "",
    schema_name_resolver: Optional[Callable] = None,
    **kwargs
):
    """
    aiohttp-apispec extension.

    Usage:

    .. code-block:: python

        from aiohttp_apispec import docs, request_schema, setup_aiohttp_apispec
        from aiohttp import web
        from marshmallow import Schema, fields


        class RequestSchema(Schema):
            id = fields.Int()
            name = fields.Str(description='name')
            bool_field = fields.Bool()


        @docs(tags=['mytag'],
              summary='Test method summary',
              description='Test method description')
        @request_schema(RequestSchema)
        async def index(request):
            return web.json_response({'msg': 'done', 'data': {}})


        app = web.Application()
        app.router.add_post('/v1/test', index)

        # init docs with all parameters, usual for ApiSpec
        setup_aiohttp_apispec(app=app,
                              title='My Documentation',
                              version='v1',
                              url='/api/docs/api-docs')

        # now we can find it on 'http://localhost:8080/api/docs/api-docs'
        web.run_app(app)

    :param Application app: aiohttp web app
    :param str title: API title
    :param str version: API version
    :param str url: url for swagger spec in JSON format
    :param str request_data_name: name of the key in Request object
                                  where validated data will be placed by
                                  validation_middleware (``'data'`` by default)
    :param error_callback: custom error handler
    :param in_place: register all routes at the moment of calling this function
                     instead of the moment of the on_startup signal.
                     If True, be sure all routes are added to router
    :param prefix: prefix to add to all registered routes
    :param schema_name_resolver: custom schema_name_resolver for MarshmallowPlugin.
    :param kwargs: any apispec.APISpec kwargs
    """
    try:
        from .openapi import AiohttpApiSpec, resolver
    except ImportError:
        logger.warning(
            "disabling openapi spec generation as 'apispec' and or 'webargs' are not installed"
        )
        return None

    if schema_name_resolver is None:
        schema_name_resolver = resolver

    return AiohttpApiSpec(
        url,
        app,
        request_data_name,
        title=title,
        version=version,
        error_callback=error_callback,
        in_place=in_place,
        prefix=prefix,
        schema_name_resolver=schema_name_resolver,
        **kwargs
    )



from string import Formatter


def get_path(route):
    path_info = route.resource.get_info()
    return path_info.get("path") or path_info.get("formatter")


def get_path_keys(path):
    return [i[1] for i in Formatter().parse(path) if i[1]]


def issubclass_py37fix(cls, cls_info):
    try:
        return issubclass(cls, cls_info)
    except TypeError:
        return False



import json
from datetime import date, datetime
from typing import Optional, Union

import asyncpg

from abrechnung.application import (
    Application,
    NotFoundError,
    check_group_permissions,
    InvalidCommand,
    create_group_log,
)
from abrechnung.domain.transactions import (
    Transaction,
    TransactionDetails,
    TransactionPosition,
    FileAttachment,
)
from abrechnung.util import parse_postgres_datetime


class TransactionService(Application):
    @staticmethod
    async def _check_transaction_permissions(
        conn: asyncpg.Connection,
        user_id: int,
        transaction_id: int,
        can_write: bool = False,
        transaction_type: Optional[Union[str, list[str]]] = None,
    ) -> int:
        """returns group id of the transaction"""
        result = await conn.fetchrow(
            "select t.type, t.group_id, can_write, is_owner "
            "from group_membership gm join transaction t on gm.group_id = t.group_id and gm.user_id = $1 "
            "where t.id = $2",
            user_id,
            transaction_id,
        )
        if not result:
            raise NotFoundError(f"user is not a member of this group")

        if can_write and not (result["can_write"] or result["is_owner"]):
            raise PermissionError(f"user does not have write permissions")

        if transaction_type:
            type_check = (
                [transaction_type]
                if isinstance(transaction_type, str)
                else transaction_type
            )
            if result["type"] not in type_check:
                raise InvalidCommand(
                    f"Transaction type {result['type']} does not match the expected type {type_check}"
                )

        return result["group_id"]

    @staticmethod
    async def _increment_transaction_version(
        conn: asyncpg.Connection, revision_id: int
    ):
        await conn.execute(
            "update transaction_revision set version = version + 1 where id = $1",
            revision_id,
        )

    @staticmethod
    def _transaction_detail_from_db_json(db_json: dict) -> TransactionDetails:
        return TransactionDetails(
            description=db_json["description"],
            value=db_json["value"],
            currency_symbol=db_json["currency_symbol"],
            currency_conversion_rate=db_json["currency_conversion_rate"],
            deleted=db_json["deleted"],
            committed_at=None
            if db_json.get("revision_committed") is None
            else parse_postgres_datetime(db_json["revision_committed"]),
            billed_at=date.fromisoformat(db_json["billed_at"]),
            creditor_shares={
                cred["account_id"]: cred["shares"]
                for cred in db_json["creditor_shares"]
            },
            debitor_shares={
                deb["account_id"]: deb["shares"] for deb in db_json["debitor_shares"]
            },
            changed_by=db_json["changed_by"],
        )

    @staticmethod
    def _transaction_position_from_db_row_json(db_json: dict) -> TransactionPosition:
        return TransactionPosition(
            id=db_json["item_id"],
            name=db_json["name"],
            price=db_json["price"],
            communist_shares=db_json["communist_shares"],
            deleted=db_json["deleted"],
            usages={
                usage["account_id"]: usage["share_amount"]
                for usage in db_json["usages"]
            },
        )

    def _file_attachment_from_db_row_json(self, db_json: dict) -> FileAttachment:
        return FileAttachment(
            id=db_json["file_id"],
            filename=db_json["filename"],
            blob_id=db_json["blob_id"],
            deleted=db_json["deleted"],
            mime_type=db_json["mime_type"],
            host_url=self.cfg["service"]["api_url"],
        )

    def _transaction_db_row(self, transaction: asyncpg.Record) -> Transaction:
        committed_details = (
            self._transaction_detail_from_db_json(
                json.loads(transaction["committed_details"])[0]
            )
            if transaction["committed_details"]
            else None
        )
        pending_details = (
            self._transaction_detail_from_db_json(
                json.loads(transaction["pending_details"])[0]
            )
            if transaction["pending_details"]
            else None
        )
        committed_positions = (
            [
                self._transaction_position_from_db_row_json(position)
                for position in json.loads(transaction["committed_positions"])
            ]
            if transaction["committed_positions"]
            else None
        )

        pending_positions = (
            [
                self._transaction_position_from_db_row_json(position)
                for position in json.loads(transaction["pending_positions"])
            ]
            if transaction["pending_positions"]
            else None
        )

        committed_files = (
            [
                self._file_attachment_from_db_row_json(file)
                for file in json.loads(transaction["committed_files"])
            ]
            if transaction["committed_files"]
            else None
        )
        pending_files = (
            [
                self._file_attachment_from_db_row_json(file)
                for file in json.loads(transaction["pending_files"])
            ]
            if transaction["pending_files"]
            else None
        )

        return Transaction(
            id=transaction["transaction_id"],
            group_id=transaction["group_id"],
            type=transaction["type"],
            is_wip=transaction["is_wip"],
            last_changed=transaction["last_changed"],
            version=transaction["version"],
            committed_details=committed_details,
            pending_details=pending_details,
            committed_positions=committed_positions,
            pending_positions=pending_positions,
            committed_files=committed_files,
            pending_files=pending_files,
        )

    async def list_transactions(
        self,
        *,
        user_id: int,
        group_id: int,
        min_last_changed: Optional[datetime] = None,
        additional_transactions: Optional[list[int]] = None,
    ) -> list[Transaction]:
        async with self.db_pool.acquire() as conn:
            async with conn.transaction():
                await check_group_permissions(
                    conn=conn, group_id=group_id, user_id=user_id
                )

                if min_last_changed:
                    # if a minimum last changed value is specified we must also return all transactions the current
                    # user has pending changes with to properly sync state across different devices of the user
                    cur = conn.cursor(
                        "select transaction_id, group_id, type, last_changed, version, is_wip, committed_details, pending_details, "
                        "   committed_positions, pending_positions, committed_files, pending_files "
                        "from full_transaction_state_valid_at($1) "
                        "where group_id = $2 "
                        "   and (is_wip or last_changed is not null and last_changed >= $3"
                        "       or (($4::int[]) is not null and transaction_id = any($4::int[])))",
                        user_id,
                        group_id,
                        min_last_changed,
                        additional_transactions,
                    )
                else:
                    cur = conn.cursor(
                        "select transaction_id, group_id, type, last_changed, version, is_wip, committed_details, pending_details, "
                        "   committed_positions, pending_positions, committed_files, pending_files "
                        "from full_transaction_state_valid_at($1) "
                        "where group_id = $2",
                        user_id,
                        group_id,
                    )

                result = []
                async for transaction in cur:
                    result.append(self._transaction_db_row(transaction))

                return result

    async def get_transaction(
        self, *, user_id: int, transaction_id: int
    ) -> Transaction:
        async with self.db_pool.acquire() as conn:
            group_id = await self._check_transaction_permissions(
                conn=conn, user_id=user_id, transaction_id=transaction_id
            )
            committed_transaction = await conn.fetchrow(
                "select transaction_id, group_id, type, last_changed, version, is_wip, committed_details, pending_details, "
                "   committed_positions, pending_positions, committed_files, pending_files "
                "from full_transaction_state_valid_at($1) "
                "where group_id = $1 and transaction_id = $2",
                group_id,
                transaction_id,
            )
            return self._transaction_db_row(committed_transaction)

    async def create_transaction(
        self,
        *,
        user_id: int,
        group_id: int,
        type: str,
        description: str,
        billed_at: date,
        currency_symbol: str,
        currency_conversion_rate: float,
        value: float,
    ) -> int:
        async with self.db_pool.acquire() as conn:
            async with conn.transaction():
                await check_group_permissions(
                    conn=conn, group_id=group_id, user_id=user_id
                )
                transaction_id = await conn.fetchval(
                    "insert into transaction (group_id, type) values ($1, $2) returning id",
                    group_id,
                    type,
                )
                revision_id = await conn.fetchval(
                    "insert into transaction_revision (user_id, transaction_id) "
                    "values ($1, $2) returning id",
                    user_id,
                    transaction_id,
                )
                revision_started = await conn.fetchval(
                    "select started from transaction_revision where id = $1",
                    revision_id,
                )
                print(f"Started after insert: {revision_started}")
                await conn.execute(
                    "insert into transaction_history (id, revision_id, currency_symbol, currency_conversion_rate, value, description, billed_at) "
                    "values ($1, $2, $3, $4, $5, $6, $7)",
                    transaction_id,
                    revision_id,
                    currency_symbol,
                    currency_conversion_rate,
                    value,
                    description,
                    billed_at,
                )
                return transaction_id

    async def commit_transaction(self, *, user_id: int, transaction_id: int) -> None:
        async with self.db_pool.acquire() as conn:
            async with conn.transaction():
                group_id = await self._check_transaction_permissions(
                    conn=conn, user_id=user_id, transaction_id=transaction_id
                )
                revision_id = await conn.fetchval(
                    "select id from transaction_revision "
                    "where transaction_id = $1 and user_id = $2 and committed is null",
                    transaction_id,
                    user_id,
                )
                if revision_id is None:
                    raise InvalidCommand(
                        f"Cannot commit a transaction without pending changes"
                    )

                await conn.execute(
                    "update transaction_revision "
                    "set committed = now(), version = version + 1 where id = $1",
                    revision_id,
                )
                await create_group_log(
                    conn=conn,
                    group_id=group_id,
                    user_id=user_id,
                    type="transaction-committed",
                    message=f"updated transaction with id {transaction_id}",
                )

    async def upload_file(
        self,
        *,
        user_id: int,
        transaction_id: int,
        filename: str,
        mime_type: str,
        content: bytes,
    ) -> int:
        # check mime type of content
        allowed_filetypes = ["image/jpeg", "image/png", "image/bmp", "image/webp"]

        if mime_type not in allowed_filetypes:
            raise InvalidCommand(f"File type {mime_type} is not an accepted file type")

        # TODO: image resizing?
        max_file_size = self.cfg["api"]["max_uploadable_file_size"]
        if len(content) / 1024 > max_file_size:
            raise InvalidCommand(f"File is too large, maximum is {max_file_size}KB")

        if "." in filename:
            raise InvalidCommand(f"Dots '.' are not allowed in file names")

        async with self.db_pool.acquire() as conn:
            async with conn.transaction():
                await self._check_transaction_permissions(
                    conn=conn,
                    user_id=user_id,
                    transaction_id=transaction_id,
                    can_write=True,
                )
                revision_id = await self._get_or_create_revision(
                    conn=conn, user_id=user_id, transaction_id=transaction_id
                )

                blob_id = await conn.fetchval(
                    "insert into blob (content, mime_type) values ($1, $2) returning id",
                    content,
                    mime_type,
                )
                file_id = await conn.fetchval(
                    "insert into file (transaction_id) values ($1) returning id",
                    transaction_id,
                )
                await conn.execute(
                    "insert into file_history (id, revision_id, filename, blob_id) values ($1, $2, $3, $4)",
                    file_id,
                    revision_id,
                    filename,
                    blob_id,
                )
                await self._increment_transaction_version(
                    conn=conn, revision_id=revision_id
                )
                return file_id

    async def delete_file(self, *, user_id: int, file_id: int) -> tuple[int, int]:
        async with self.db_pool.acquire() as conn:
            async with conn.transaction():
                perms = await conn.fetchrow(
                    "select t.id as transaction_id "
                    "from group_membership gm "
                    "   join transaction t on gm.group_id = t.group_id and gm.user_id = $1 "
                    "   join file f on t.id = f.transaction_id "
                    "where f.id = $2 and gm.can_write",
                    user_id,
                    file_id,
                )
                if not perms:
                    raise InvalidCommand("File not found")

                transaction_id = perms["transaction_id"]
                committed_state = await conn.fetchrow(
                    "select filename, deleted from committed_file_state_valid_at() where file_id = $1",
                    file_id,
                )
                if committed_state is not None and committed_state["deleted"]:
                    raise InvalidCommand("Cannot delete file as it is already deleted")

                if committed_state is None:
                    # file is only attached to a pending change, fully delete it right away, blob will be cleaned up
                    pending_state = await conn.fetchrow(
                        "select revision_id from aggregated_pending_file_history "
                        "where file_id = $1 and changed_by = $2",
                        file_id,
                        user_id,
                    )
                    if pending_state is None:
                        raise InvalidCommand("Unknown error occurred")

                    revision_id = pending_state["revision_id"]

                    await conn.execute(
                        "update file_history fh set deleted = true, blob_id = null where id = $1 and revision_id = $2",
                        file_id,
                        revision_id,
                    )
                    await self._increment_transaction_version(
                        conn=conn, revision_id=revision_id
                    )
                    return transaction_id, pending_state["revision_id"]

                revision_id = await self._get_or_create_revision(
                    conn=conn, user_id=user_id, transaction_id=transaction_id
                )

                await conn.execute(
                    "insert into file_history(id, revision_id, filename, blob_id, deleted) "
                    "values ($1, $2, $3, null, true)",
                    file_id,
                    revision_id,
                    committed_state["filename"],
                )
                await self._increment_transaction_version(
                    conn=conn, revision_id=revision_id
                )
                return transaction_id, revision_id

    async def read_file_contents(
        self, user_id: int, file_id: int, blob_id: int
    ) -> tuple[str, bytes]:
        async with self.db_pool.acquire() as conn:
            perms = await conn.fetchrow(
                "select f.id "
                "from group_membership gm "
                "   join transaction t on gm.group_id = t.group_id and gm.user_id = $1 "
                "   join file f on t.id = f.transaction_id and f.id = $2"
                "   join file_history fh on f.id = fh.id "
                "where fh.blob_id = $3",
                user_id,
                file_id,
                blob_id,
            )
            if not perms:
                raise InvalidCommand("File not found")

            blob = await conn.fetchrow(
                "select content, mime_type from blob where id = $1", blob_id
            )
            if not blob:
                raise InvalidCommand("File not found")

            return blob["mime_type"], blob["content"]

    async def update_transaction(
        self,
        *,
        user_id: int,
        transaction_id: int,
        value: float,
        description: str,
        billed_at: date,
        currency_symbol: str,
        currency_conversion_rate: float,
    ):
        async with self.db_pool.acquire() as conn:
            async with conn.transaction():
                await self._check_transaction_permissions(
                    conn=conn,
                    user_id=user_id,
                    transaction_id=transaction_id,
                    can_write=True,
                )
                revision_id = await self._get_or_create_pending_transaction_change(
                    conn=conn, user_id=user_id, transaction_id=transaction_id
                )
                await conn.execute(
                    "update transaction_history th "
                    "set value = $3, description = $4, currency_symbol = $5, currency_conversion_rate = $6, billed_at = $7 "
                    "where th.id = $1 and th.revision_id = $2",
                    transaction_id,
                    revision_id,
                    value,
                    description,
                    currency_symbol,
                    currency_conversion_rate,
                    billed_at,
                )
                await self._increment_transaction_version(
                    conn=conn, revision_id=revision_id
                )

    async def create_transaction_change(self, *, user_id: int, transaction_id: int):
        async with self.db_pool.acquire() as conn:
            async with conn.transaction():
                await self._check_transaction_permissions(
                    conn=conn,
                    user_id=user_id,
                    transaction_id=transaction_id,
                    can_write=True,
                )
                await self._get_or_create_revision(
                    conn=conn, user_id=user_id, transaction_id=transaction_id
                )

    async def discard_transaction_changes(self, *, user_id: int, transaction_id: int):
        async with self.db_pool.acquire() as conn:
            async with conn.transaction():
                await self._check_transaction_permissions(
                    conn=conn,
                    user_id=user_id,
                    transaction_id=transaction_id,
                    can_write=True,
                )

                revision_id = await conn.fetchval(
                    "select id "
                    "from transaction_revision tr where tr.user_id = $1 and tr.transaction_id = $2 "
                    "   and tr.committed is null",
                    user_id,
                    transaction_id,
                )
                if revision_id is None:
                    raise InvalidCommand(
                        f"No changes to discard for transaction {transaction_id}"
                    )

                last_committed_revision = await conn.fetchval(
                    "select id "
                    "from transaction_revision tr where tr.transaction_id = $1 and tr.committed is not null",
                    transaction_id,
                )
                if (
                    last_committed_revision is None
                ):  # we have a newly created transaction - disallow discarding changes
                    raise InvalidCommand(
                        f"Cannot discard transaction changes without any committed changes"
                    )
                else:
                    await conn.execute(
                        "delete from transaction_revision tr " "where tr.id = $1",
                        revision_id,
                    )

    async def delete_transaction(self, *, user_id: int, transaction_id: int):
        async with self.db_pool.acquire() as conn:
            async with conn.transaction():
                group_id = await self._check_transaction_permissions(
                    conn=conn,
                    user_id=user_id,
                    transaction_id=transaction_id,
                    can_write=True,
                )

                row = await conn.fetchrow(
                    "select description, revision_id, deleted "
                    "from committed_transaction_state_valid_at() "
                    "where transaction_id = $1",
                    transaction_id,
                )
                if row is not None and row["deleted"]:
                    raise InvalidCommand(
                        f"Cannot delete transaction {transaction_id} as it already is deleted"
                    )

                await create_group_log(
                    conn=conn,
                    group_id=group_id,
                    user_id=user_id,
                    type="transaction-deleted",
                    message=f"deleted transaction with id {transaction_id}",
                )

                if (
                    row is None
                ):  # the transaction has no committed changes, we can only delete it if we created it
                    revision_id = await conn.fetchval(
                        "select id from transaction_revision tr "
                        "where tr.user_id = $1 and tr.transaction_id = $2 and tr.committed is null",
                        user_id,
                        transaction_id,
                    )
                    if revision_id is None:
                        raise InvalidCommand(
                            f"Cannot delete uncommitted transaction {transaction_id} of another user"
                        )

                    # here we assume there has already been a transaction_history entry, if not something weird has
                    # happened
                    t_id = await conn.fetchval(
                        "update transaction_history th set deleted = true "
                        "where th.id = $1 and th.revision_id = $2 returning id",
                        transaction_id,
                        revision_id,
                    )
                    if t_id is None:
                        raise InvalidCommand(
                            f"something weird has happened deleting uncommitted transaction "
                            f"{transaction_id}, please consult your local IT admin"
                        )

                    # now commit this change
                    await conn.execute(
                        "update transaction_revision tr set committed = now() "
                        "where tr.user_id = $1 and tr.transaction_id = $2 and tr.committed is null",
                        user_id,
                        transaction_id,
                    )
                    return

                else:  # we have at least one committed change for this transaction
                    revision_id = await self._get_or_create_pending_transaction_change(
                        conn=conn, user_id=user_id, transaction_id=transaction_id
                    )

                    await conn.execute(
                        "update transaction_history th set deleted = true "
                        "where th.id = $1 and th.revision_id = $2",
                        transaction_id,
                        revision_id,
                    )

                    await conn.execute(
                        "update transaction_revision tr set committed = NOW() "
                        "where tr.id = $1",
                        revision_id,
                    )

    async def _get_or_create_revision(
        self, conn: asyncpg.Connection, user_id: int, transaction_id: int
    ) -> int:
        """return the revision id, assumes we are already in a transaction"""
        revision_id = await conn.fetchval(
            "select id "
            "from transaction_revision "
            "where transaction_id = $1 and user_id = $2 and committed is null",
            transaction_id,
            user_id,
        )
        if revision_id:  # there already is a wip revision
            return revision_id

        # create a new transaction revision
        revision_id = await conn.fetchval(
            "insert into transaction_revision (user_id, transaction_id) values ($1, $2) returning id",
            user_id,
            transaction_id,
        )
        return revision_id

    async def _get_or_create_pending_transaction_change(
        self, conn: asyncpg.Connection, user_id: int, transaction_id: int
    ) -> int:
        revision_id = await self._get_or_create_revision(
            conn=conn, user_id=user_id, transaction_id=transaction_id
        )

        t = await conn.fetchval(
            "select id from transaction_history th where revision_id = $1 and id = $2",
            revision_id,
            transaction_id,
        )
        if t:
            return revision_id

        last_committed_revision = await conn.fetchval(
            "select tr.id "
            "from transaction_revision tr "
            "   join transaction_history th on tr.id = th.revision_id and tr.transaction_id = th.id "
            "where tr.transaction_id = $1 and tr.committed is not null "
            "order by tr.committed desc "
            "limit 1",
            transaction_id,
        )

        if last_committed_revision is None:
            raise InvalidCommand(
                f"Cannot edit transaction {transaction_id} as it has no committed changes."
            )

        # copy all existing transaction data into a new history entry
        await conn.execute(
            "insert into transaction_history (id, revision_id, currency_symbol, currency_conversion_rate, description, value, billed_at, deleted)"
            "select id, $1, currency_symbol, currency_conversion_rate, description, value, billed_at, deleted "
            "from transaction_history where id = $2 and revision_id = $3",
            revision_id,
            transaction_id,
            last_committed_revision,
        )

        # copy all last committed creditor shares
        await conn.execute(
            "insert into creditor_share (transaction_id, revision_id, account_id, shares) "
            "select transaction_id, $1, account_id, shares "
            "from creditor_share where transaction_id = $2 and revision_id = $3",
            revision_id,
            transaction_id,
            last_committed_revision,
        )

        # copy all last committed debitor shares
        await conn.execute(
            "insert into debitor_share (transaction_id, revision_id, account_id, shares) "
            "select transaction_id, $1, account_id, shares "
            "from debitor_share where transaction_id = $2 and revision_id = $3",
            revision_id,
            transaction_id,
            last_committed_revision,
        )

        return revision_id

    async def _get_or_create_pending_purchase_item_change(
        self, conn: asyncpg.Connection, user_id: int, item_id: int
    ):
        purchase_item = await conn.fetchrow(
            "select id, transaction_id from purchase_item pi where id = $1", item_id
        )
        if not purchase_item:
            raise RuntimeError(f"Item with id {item_id} does not exist")
        revision_id = await self._get_or_create_revision(
            conn=conn, user_id=user_id, transaction_id=purchase_item["transaction_id"]
        )

        history_entry = await conn.fetchval(
            "select id from purchase_item_history pih where pih.id = $1 and pih.revision_id = $2",
            item_id,
            revision_id,
        )
        if history_entry:
            return revision_id

        last_committed_revision = await conn.fetchval(
            "select tr.id "
            "from transaction_revision tr "
            "   join purchase_item pi on pi.transaction_id = tr.transaction_id "
            "   join purchase_item_history pih on pih.revision_id = tr.id and  pi.id = pih.id "
            "where tr.transaction_id = $1 and tr.committed is not null "
            "order by tr.committed desc "
            "limit 1",
            purchase_item["transaction_id"],
        )

        await conn.execute(
            "insert into purchase_item_history (id, revision_id, name, price, communist_shares, deleted) "
            "select $1, $2, pih.name, pih.price, pih.communist_shares, pih.deleted "
            "from purchase_item_history pih join purchase_item pi on pih.id = pi.id "
            "where pi.id = $1 and pi.transaction_id = $3 and pih.revision_id = $4 and not pih.deleted",
            item_id,
            revision_id,
            purchase_item["transaction_id"],
            last_committed_revision,
        )
        await conn.execute(
            "insert into purchase_item_usage (item_id, revision_id, account_id, share_amount) "
            "select $1, $2, account_id, share_amount "
            "from purchase_item_usage piu join purchase_item pi on piu.item_id = pi.id "
            "join purchase_item_history pih on pi.id = pih.id and pih.revision_id = piu.revision_id "
            "where pi.id = $1 and pi.transaction_id = $3 and piu.revision_id = $4 and not pih.deleted",
            item_id,
            revision_id,
            purchase_item["transaction_id"],
            last_committed_revision,
        )

        return revision_id

    async def _transaction_share_check(
        self,
        conn: asyncpg.Connection,
        user_id: int,
        transaction_id: int,
        account_id: int,
        transaction_type: Optional[Union[str, list[str]]] = None,
    ) -> tuple[int, int]:
        """returns tuple of group_id of the transaction and the users revision_id of the pending change"""
        group_id = await self._check_transaction_permissions(
            conn=conn,
            user_id=user_id,
            transaction_id=transaction_id,
            can_write=True,
            transaction_type=transaction_type,
        )
        revision_id = await self._get_or_create_pending_transaction_change(
            conn=conn, user_id=user_id, transaction_id=transaction_id
        )

        acc = await conn.fetchval(
            "select id from account where group_id = $1", group_id
        )
        if not acc:
            raise NotFoundError(f"Account with id {account_id}")

        return group_id, revision_id

    async def add_or_change_creditor_share(
        self,
        *,
        user_id: int,
        transaction_id: int,
        account_id: int,
        value: float,
    ):
        async with self.db_pool.acquire() as conn:
            async with conn.transaction():
                group_id, revision_id = await self._transaction_share_check(
                    conn, user_id, transaction_id, account_id
                )

                await conn.execute(
                    "insert into creditor_share (transaction_id, revision_id, account_id, shares) "
                    "values ($1, $2, $3, $4) "
                    "on conflict (transaction_id, revision_id, account_id) do "
                    "update set shares = $4 "
                    "where creditor_share.transaction_id = $1 "
                    "   and creditor_share.revision_id = $2 "
                    "   and creditor_share.account_id = $3",
                    transaction_id,
                    revision_id,
                    account_id,
                    value,
                )
                await self._increment_transaction_version(
                    conn=conn, revision_id=revision_id
                )

    async def switch_creditor_share(
        self,
        *,
        user_id: int,
        transaction_id: int,
        account_id: int,
        value: float,
    ):
        async with self.db_pool.acquire() as conn:
            async with conn.transaction():
                group_id, revision_id = await self._transaction_share_check(
                    conn,
                    user_id,
                    transaction_id,
                    account_id,
                    transaction_type=["purchase", "transfer"],
                )

                await conn.execute(
                    "delete from creditor_share where transaction_id = $1 and revision_id = $2",
                    transaction_id,
                    revision_id,
                )
                await conn.execute(
                    "insert into creditor_share (transaction_id, revision_id, account_id, shares) "
                    "values ($1, $2, $3, $4)",
                    transaction_id,
                    revision_id,
                    account_id,
                    value,
                )
                await self._increment_transaction_version(
                    conn=conn, revision_id=revision_id
                )

    async def delete_creditor_share(
        self, *, user_id: int, transaction_id: int, account_id: int
    ):
        async with self.db_pool.acquire() as conn:
            async with conn.transaction():
                group_id, revision_id = await self._transaction_share_check(
                    conn, user_id, transaction_id, account_id
                )

                r = await conn.fetchval(
                    "delete from creditor_share where transaction_id = $1 and revision_id = $2 and account_id = $3 "
                    "returning revision_id",
                    transaction_id,
                    revision_id,
                    account_id,
                )
                if not r:
                    raise NotFoundError(f"Creditor share does not exist")

                await self._increment_transaction_version(
                    conn=conn, revision_id=revision_id
                )

    async def add_or_change_debitor_share(
        self,
        *,
        user_id: int,
        transaction_id: int,
        account_id: int,
        value: float,
    ):
        async with self.db_pool.acquire() as conn:
            async with conn.transaction():
                group_id, revision_id = await self._transaction_share_check(
                    conn, user_id, transaction_id, account_id
                )

                await conn.execute(
                    "insert into debitor_share (transaction_id, revision_id, account_id, shares) "
                    "values ($1, $2, $3, $4) "
                    "on conflict (transaction_id, revision_id, account_id) do "
                    "update set shares = $4 "
                    "where debitor_share.transaction_id = $1 "
                    "   and debitor_share.revision_id = $2 "
                    "   and debitor_share.account_id = $3",
                    transaction_id,
                    revision_id,
                    account_id,
                    value,
                )
                await self._increment_transaction_version(
                    conn=conn, revision_id=revision_id
                )

    async def switch_debitor_share(
        self, *, user_id: int, transaction_id: int, account_id: int, value: float
    ):
        async with self.db_pool.acquire() as conn:
            async with conn.transaction():
                group_id, revision_id = await self._transaction_share_check(
                    conn,
                    user_id,
                    transaction_id,
                    account_id,
                    transaction_type=["transfer"],
                )

                await conn.execute(
                    "delete from debitor_share where transaction_id = $1 and revision_id = $2",
                    transaction_id,
                    revision_id,
                )
                await conn.execute(
                    "insert into debitor_share (transaction_id, revision_id, account_id, shares) "
                    "values ($1, $2, $3, $4)",
                    transaction_id,
                    revision_id,
                    account_id,
                    value,
                )
                await self._increment_transaction_version(
                    conn=conn, revision_id=revision_id
                )

    async def delete_debitor_share(
        self, *, user_id: int, transaction_id: int, account_id: int
    ):
        async with self.db_pool.acquire() as conn:
            async with conn.transaction():
                group_id, revision_id = await self._transaction_share_check(
                    conn, user_id, transaction_id, account_id
                )

                r = await conn.fetchval(
                    "delete from debitor_share where transaction_id = $1 and revision_id = $2 and account_id = $3 "
                    "returning revision_id",
                    transaction_id,
                    revision_id,
                    account_id,
                )
                if not r:
                    raise NotFoundError(f"Debitor share does not exist")

                await self._increment_transaction_version(
                    conn=conn, revision_id=revision_id
                )

    async def create_purchase_item(
        self,
        *,
        user_id: int,
        transaction_id: int,
        name: str,
        price: float,
        communist_shares: float,
        usages: Optional[dict[int, float]] = None,
    ) -> int:
        async with self.db_pool.acquire() as conn:
            async with conn.transaction():
                group_id = await self._check_transaction_permissions(
                    conn=conn,
                    user_id=user_id,
                    transaction_id=transaction_id,
                    can_write=True,
                    transaction_type="purchase",
                )
                revision_id = await self._get_or_create_revision(
                    conn=conn, user_id=user_id, transaction_id=transaction_id
                )
                item_id = await conn.fetchval(
                    "insert into purchase_item (transaction_id) values ($1) returning id",
                    transaction_id,
                )
                await conn.execute(
                    "insert into purchase_item_history (id, revision_id, name, price, communist_shares) "
                    "values ($1, $2, $3, $4, $5)",
                    item_id,
                    revision_id,
                    name,
                    price,
                    communist_shares,
                )
                if usages:
                    if len(usages.keys()) != len(set(usages.keys())):
                        raise InvalidCommand(
                            f"an account cannot appear twice in position usages"
                        )

                    n_accounts = await conn.fetchval(
                        "select count(*) from account where group_id = $1 and id = any($2::int[])",
                        group_id,
                        list(usages.keys()),
                    )
                    if len(usages.keys()) != n_accounts:
                        raise InvalidCommand(
                            "one of the given accounts does not exist in this group"
                        )

                    for account_id, share_amount in usages.items():
                        await conn.execute(
                            "insert into purchase_item_usage (item_id, revision_id, account_id, share_amount) "
                            "values ($1, $2, $3, $4)",
                            item_id,
                            revision_id,
                            account_id,
                            share_amount,
                        )

                await self._increment_transaction_version(
                    conn=conn, revision_id=revision_id
                )
                return item_id

    @staticmethod
    async def _check_purchase_item_permissions(
        *, conn: asyncpg.Connection, user_id: int, item_id: int, can_write: bool = True
    ) -> tuple[int, int]:
        row = await conn.fetchrow(
            "select gm.group_id as group_id, gm.can_write as can_write, t.id as transaction_id "
            "from group_membership gm "
            "join transaction t on gm.group_id = t.group_id "
            "join purchase_item pi on t.id = pi.transaction_id "
            "where pi.id = $1 and gm.user_id = $2",
            item_id,
            user_id,
        )
        if row is None:
            raise NotFoundError(f"A purchase item with id '{item_id}' does not exist")

        if can_write and not row["can_write"]:
            raise PermissionError(
                f"Missing write permissions for item with id '{item_id}'"
            )

        return row["group_id"], row["transaction_id"]

    async def update_purchase_item(
        self,
        *,
        user_id: int,
        item_id: int,
        name: str,
        price: float,
        communist_shares: float,
    ) -> tuple[int, int]:
        """Returns [transaction_id, revision_id]"""
        async with self.db_pool.acquire() as conn:
            async with conn.transaction():
                group_id, transaction_id = await self._check_purchase_item_permissions(
                    conn=conn, user_id=user_id, item_id=item_id
                )
                revision_id = await self._get_or_create_pending_purchase_item_change(
                    conn=conn, user_id=user_id, item_id=item_id
                )
                await conn.execute(
                    "update purchase_item_history set name = $3, price = $4, communist_shares = $5 "
                    "where id = $1 and revision_id = $2",
                    item_id,
                    revision_id,
                    name,
                    price,
                    communist_shares,
                )

                await self._increment_transaction_version(
                    conn=conn, revision_id=revision_id
                )
                return transaction_id, revision_id

    async def add_or_change_item_share(
        self,
        *,
        user_id: int,
        item_id: int,
        account_id: int,
        share_amount: float,
    ) -> tuple[int, int]:
        """Returns [transaction_id, revision_id]"""
        async with self.db_pool.acquire() as conn:
            async with conn.transaction():
                _, transaction_id = await self._check_purchase_item_permissions(
                    conn=conn, user_id=user_id, item_id=item_id
                )
                revision_id = await self._get_or_create_pending_purchase_item_change(
                    conn=conn, user_id=user_id, item_id=item_id
                )

                await conn.execute(
                    "insert into purchase_item_usage (item_id, revision_id, account_id, share_amount) "
                    "values ($1, $2, $3, $4) "
                    "on conflict (item_id, revision_id, account_id) do "
                    "update set share_amount = $4 "
                    "where purchase_item_usage.item_id = $1 "
                    "   and purchase_item_usage.revision_id = $2 "
                    "   and purchase_item_usage.account_id = $3",
                    item_id,
                    revision_id,
                    account_id,
                    share_amount,
                )
                await self._increment_transaction_version(
                    conn=conn, revision_id=revision_id
                )
                return transaction_id, revision_id

    async def delete_item_share(
        self,
        *,
        user_id: int,
        item_id: int,
        account_id: int,
    ) -> tuple[int, int]:
        """Returns [transaction_id, revision_id]"""
        async with self.db_pool.acquire() as conn:
            async with conn.transaction():
                _, transaction_id = await self._check_purchase_item_permissions(
                    conn=conn, user_id=user_id, item_id=item_id
                )
                revision_id = await self._get_or_create_pending_purchase_item_change(
                    conn=conn, user_id=user_id, item_id=item_id
                )

                r = await conn.fetchval(
                    "delete from purchase_item_usage where item_id = $1 and revision_id = $2 and account_id = $3 "
                    "returning revision_id",
                    item_id,
                    revision_id,
                    account_id,
                )
                if not r:
                    raise NotFoundError(f"Purchase item usage does not exist")

                await self._increment_transaction_version(
                    conn=conn, revision_id=revision_id
                )
                return transaction_id, revision_id

    async def delete_purchase_item(
        self, *, user_id: int, item_id: int
    ) -> tuple[int, int]:
        """Returns [transaction_id, revision_id]"""
        async with self.db_pool.acquire() as conn:
            async with conn.transaction():
                _, transaction_id = await self._check_purchase_item_permissions(
                    conn=conn, user_id=user_id, item_id=item_id
                )
                revision_id = await self._get_or_create_pending_purchase_item_change(
                    conn=conn, user_id=user_id, item_id=item_id
                )

                await conn.execute(
                    "update purchase_item_history pih set deleted = true "
                    "where pih.id = $1 and pih.revision_id = $2",
                    item_id,
                    revision_id,
                )

                await self._increment_transaction_version(
                    conn=conn, revision_id=revision_id
                )
                return transaction_id, revision_id



import asyncio
import logging

from aiohttp import web
from aiohttp.test_utils import (
    TestServer,
    TestClient,
)

from abrechnung.application.accounts import AccountService
from abrechnung.application.groups import GroupService
from abrechnung.application.transactions import TransactionService
from abrechnung.application.users import UserService
from abrechnung.config import Config
from abrechnung.http import HTTPService
from abrechnung.http.auth import token_for_user
from tests import AsyncTestCase


class AsyncHTTPTestCase(AsyncTestCase):
    async def get_application(self) -> web.Application:
        """
        This method should be overridden
        to return the aiohttp.web.Application
        object to test.

        """
        return self.get_app()

    def get_app(self) -> web.Application:
        """Obsolete method used to constructing web application.

        Use .get_application() coroutine instead

        """
        raise RuntimeError("Did you forget to define get_application()?")

    def setUp(self) -> None:
        try:
            self.loop = asyncio.get_running_loop()
        except (AttributeError, RuntimeError):  # AttributeError->py36
            self.loop = asyncio.get_event_loop_policy().get_event_loop()

        self.loop.run_until_complete(self._setup_db())
        self.loop.run_until_complete(self.setUpAsync())

    async def setUpAsync(self) -> None:
        self.app = await self.get_application()
        self.server = await self.get_server(self.app)
        self.client = await self.get_client(self.server)

        await self.client.start_server()

    def tearDown(self) -> None:
        self.loop.run_until_complete(self.tearDownAsync())
        self.loop.run_until_complete(self._teardown_db())

    async def tearDownAsync(self) -> None:
        await self.client.close()

    async def get_server(self, app: web.Application) -> TestServer:
        """Return a TestServer instance."""
        return TestServer(app, loop=self.loop)

    async def get_client(self, server: TestServer) -> TestClient:
        """Return a TestClient instance."""
        return TestClient(server, loop=self.loop)


class BaseHTTPAPITest(AsyncHTTPTestCase):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        logging.basicConfig(level=logging.DEBUG)

    async def tearDownAsync(self) -> None:
        await super().tearDownAsync()
        await self.http_service._unregister_forwarder(
            self.db_conn, forwarder_id="test_forwarder"
        )

    async def get_application(self) -> web.Application:
        self.secret_key = "asdf1234"
        self.test_config = Config.from_dict(
            {
                "api": {"secret_key": self.secret_key, "enable_registration": True},
                "service": {"api_url": "http://localhost/api"},
            }
        )
        self.http_service = HTTPService(config=self.test_config)

        await self.http_service._register_forwarder(
            self.db_conn, forwarder_id="test_forwarder"
        )

        self.group_service = GroupService(self.db_pool, config=self.test_config)
        self.account_service = AccountService(self.db_pool, config=self.test_config)
        self.user_service = UserService(self.db_pool, config=self.test_config)
        self.transaction_service = TransactionService(
            self.db_pool, config=self.test_config
        )

        app = self.http_service.create_app(db_pool=self.db_pool)

        return app


class HTTPAPITest(BaseHTTPAPITest):
    async def setUpAsync(self) -> None:
        await super().setUpAsync()

        self.test_user_id, password = await self._create_test_user(
            "user1", "user1@email.stuff"
        )
        _, session_id, _ = await self.user_service.login_user(
            "user1", password=password, session_name="session1"
        )
        self.jwt_token = token_for_user(self.test_user_id, session_id, self.secret_key)

    async def _post(self, *args, **kwargs):
        headers = kwargs.pop("headers", {})
        headers.update({"Authorization": f"Bearer {self.jwt_token}"})
        kwargs["headers"] = headers
        return await self.client.post(*args, **kwargs)

    async def _delete(self, *args, **kwargs):
        headers = kwargs.pop("headers", {})
        headers.update({"Authorization": f"Bearer {self.jwt_token}"})
        kwargs["headers"] = headers
        return await self.client.delete(*args, **kwargs)

    async def _get(self, *args, **kwargs):
        headers = kwargs.pop("headers", {})
        headers.update({"Authorization": f"Bearer {self.jwt_token}"})
        kwargs["headers"] = headers
        return await self.client.get(*args, **kwargs)



def response_schema(schema, code=200, required=False, description=None):
    """
    Add response info into the swagger spec

    Usage:

    .. code-block:: python

        from aiohttp import web
        from marshmallow import Schema, fields


        class ResponseSchema(Schema):
            msg = fields.Str()
            data = fields.Dict()

        @response_schema(ResponseSchema(), 200)
        async def index(request):
            return web.json_response({'msg': 'done', 'data': {}})

    :param str description: response description
    :param bool required:
    :param schema: :class:`Schema <marshmallow.Schema>` class or instance
    :param int code: HTTP response code
    """
    if callable(schema):
        schema = schema()

    def wrapper(func):
        if not hasattr(func, "__apispec__"):
            func.__apispec__ = {"schemas": [], "responses": {}, "parameters": []}
            func.__schemas__ = []
        func.__apispec__["responses"]["%s" % code] = {
            "schema": schema,
            "required": required,
            "description": description or "",
        }
        return func

    return wrapper


# For backward compatibility
marshal_with = response_schema
